"""Import tasks from the catalog available on VizieR.
"""
import csv
import os
from math import isnan

from astropy.time import Time as astrotime
from astroquery.vizier import Vizier

from astrocats.catalog.utils import (convert_aq_output, get_sig_digits,
                                     is_number, jd_to_mjd, make_date_string,
                                     pbar, pretty_num, rep_chars, round_sig,
                                     uniq_cdl)
from astrocats.supernovae.utils import radec_clean
from cdecimal import Decimal

from ..constants import CLIGHT, KM


def do_vizier(catalog):
    """
    """
    task_str = catalog.get_current_task_str()

    Vizier.ROW_LIMIT = -1

    # 2012ApJS..200...12H
    result = Vizier.get_catalogs('J/ApJS/200/12/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    oldname = ''
    for row in pbar(table, task_str):
        name = row['SN']
        if is_number(name[:4]):
            name = 'SN' + name
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2012ApJS..200...12H')
        catalog.entries[name].add_quantity('alias', name, source)
        if '[' not in row['Gal']:
            catalog.entries[name].add_quantity(
                'host', row['Gal'].replace('_', ' '), source)
        catalog.entries[name].add_quantity('redshift', str(
            row['z']), source, kind='heliocentric')
        catalog.entries[name].add_quantity('redshift', str(
            row['zCMB']), source, kind='cmb')
        catalog.entries[name].add_quantity('ebv', str(
            row['E_B-V_']), source, error=str(row['e_E_B-V_']) if
            row['e_E_B-V_'] else '')
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)

    # 2012ApJ...746...85S
    result = Vizier.get_catalogs('J/ApJ/746/85/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    oldname = ''
    for row in pbar(table, task_str):
        name = row['Name'].replace('SCP', 'SCP-')
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2012ApJ...746...85S')
        catalog.entries[name].add_quantity('alias', name, source)
        if row['f_Name']:
            catalog.entries[name].add_quantity('claimedtype', 'Ia', source)
        if row['z']:
            catalog.entries[name].add_quantity('redshift', str(
                row['z']), source, kind='spectroscopic')
        else:
            catalog.entries[name].add_quantity('redshift', str(
                row['zCl']), source, kind='cluster')
        catalog.entries[name].add_quantity('ebv', str(row['E_B-V_']), source)
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)

    result = Vizier.get_catalogs('J/ApJ/746/85/table2')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    oldname = ''
    for row in pbar(table, task_str):
        name = row['Name'].replace('SCP', 'SCP-')
        flux = Decimal(float(row['Flux']))
        if flux <= 0.0:
            continue
        err = Decimal(float(row['e_Flux']))
        zp = Decimal(float(row['Zero']))
        sig = get_sig_digits(str(row['Flux'])) + 1
        magnitude = pretty_num(zp - Decimal(2.5) * (flux.log10()), sig=sig)
        e_magnitude = pretty_num(
            Decimal(2.5) * (Decimal(1.0) + err / flux).log10(), sig=sig)
        if float(e_magnitude) > 5.0:
            continue
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2012ApJ...746...85S')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_photometry(
            time=str(row['MJD']), band=row['Filter'],
            instrument=row['Inst'],
            magnitude=magnitude, e_magnitude=e_magnitude, source=source)

    # 2004ApJ...602..571B
    result = Vizier.get_catalogs('J/ApJ/602/571/table8')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    oldname = ''
    for row in pbar(table, task_str):
        name = 'SN' + row['SN']
        flux = Decimal(float(row['Flux']))
        if flux <= 0.0:
            continue
        err = Decimal(float(row['e_Flux']))
        sig = get_sig_digits(str(row['Flux'])) + 1
        magnitude = pretty_num(Decimal(25.0) - Decimal(2.5) * (flux.log10()),
                               sig=sig)
        e_magnitude = pretty_num(
            Decimal(2.5) * (Decimal(1.0) + err / flux).log10(), sig=sig)
        if float(e_magnitude) > 5.0:
            continue
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2004ApJ...602..571B')
        catalog.entries[name].add_quantity('alias', name, source)
        band = row['Filt']
        system = ''
        telescope = ''
        if band in ['R', 'I']:
            system = 'Cousins'
        if band == 'Z':
            telescope = 'Subaru'
        catalog.entries[name].add_photometry(
            time=str(row['MJD']), band=band, system=system,
            telescope=telescope,
            magnitude=magnitude, e_magnitude=e_magnitude, source=source)

    # 2014MNRAS.444.3258M
    result = Vizier.get_catalogs('J/MNRAS/444/3258/SNe')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    oldname = ''
    for row in pbar(table, task_str):
        name = row['SN']
        if name == oldname:
            continue
        oldname = name
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2014MNRAS.444.3258M')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity('redshift', str(
            row['z']), source, kind='heliocentric', error=str(row['e_z']))
        catalog.entries[name].add_quantity(
            'ra', str(row['_RA']), source, unit='floatdegrees')
        catalog.entries[name].add_quantity('dec', str(
            row['_DE']), source, unit='floatdegrees')
    catalog.journal_entries()

    # 2014MNRAS.438.1391P
    result = Vizier.get_catalogs('J/MNRAS/438/1391/table2')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        name = row['SN']
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2014MNRAS.438.1391P')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity('redshift', str(
            row['zh']), source, kind='heliocentric')
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
    catalog.journal_entries()

    # 2012ApJ...749...18B
    result = Vizier.get_catalogs('J/ApJ/749/18/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        name = row['Name'].replace(' ', '')
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2012ApJ...749...18B')
        catalog.entries[name].add_quantity('alias', name, source)
        mjd = str(astrotime(2450000. + row['JD'], format='jd').mjd)
        band = row['Filt']
        magnitude = str(row['mag'])
        e_magnitude = str(row['e_mag'])
        e_magnitude = '' if e_magnitude == '--' else e_magnitude
        upperlimit = True if row['l_mag'] == '>' else False
        catalog.entries[name].add_photometry(
            time=mjd, band=band, magnitude=magnitude,
            e_magnitude=e_magnitude, instrument='UVOT',
            source=source, upperlimit=upperlimit, telescope='Swift',
            system='Swift')
    catalog.journal_entries()

    # 2010A&A...523A...7G
    result = Vizier.get_catalogs('J/A+A/523/A7/table9')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        name = 'SNLS-' + row['SNLS']
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2010A&A...523A...7G')
        catalog.entries[name].add_quantity('alias', name, source)
        astrot = astrotime(2450000. + row['Date1'], format='jd').datetime
        catalog.entries[name].add_quantity('discoverdate', make_date_string(
            astrot.year, astrot.month, astrot.day), source)
        catalog.entries[name].add_quantity('ebv', str(row['E_B-V_']), source)
        catalog.entries[name].add_quantity('redshift', str(
            row['z']), source, kind='heliocentric')
        type_str = (row['Type']
                    .replace('*', '?').replace('SN', '')
                    .replace('(pec)', ' P').replace('Ia? P?', 'Ia P?'))
        catalog.entries[name].add_quantity('claimedtype', type_str, source)
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
    catalog.journal_entries()

    # 2004A&A...415..863G
    result = Vizier.get_catalogs('J/A+A/415/863/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        name = 'SN' + row['SN']
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2004A&A...415..863G')
        catalog.entries[name].add_quantity('alias', name, source)
        datesplit = row['Date'].split('-')
        date_str = make_date_string(datesplit[0], datesplit[1].lstrip('0'),
                                    datesplit[2].lstrip('0'))
        catalog.entries[name].add_quantity('discoverdate', date_str, source)
        catalog.entries[name].add_quantity(
            'host', 'Abell ' + str(row['Abell']), source)
        catalog.entries[name].add_quantity('claimedtype', row['Type'], source)
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
        if row['zSN']:
            catalog.entries[name].add_quantity('redshift', str(
                row['zSN']), source, kind='spectroscopic')
        else:
            catalog.entries[name].add_quantity('redshift', str(
                row['zCl']), source, kind='cluster')
    catalog.journal_entries()

    # 2008AJ....136.2306H
    result = Vizier.get_catalogs('J/AJ/136/2306/sources')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        name = 'SDSS-II SN ' + str(row['SNID'])
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2008AJ....136.2306H')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity(
            'claimedtype', row['SpType'].replace('SN.', '').strip(':'), source)
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)

    # 2010ApJ...708..661D
    result = Vizier.get_catalogs('J/ApJ/708/661/sn')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        name = row['SN']
        if not name:
            name = 'SDSS-II SN ' + str(row['SDSS-II'])
        else:
            name = 'SN' + name
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2010ApJ...708..661D')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity(
            'alias', 'SDSS-II ' + str(row['SDSS-II']), source)
        catalog.entries[name].add_quantity('claimedtype', 'II P', source)
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)

    result = Vizier.get_catalogs('J/ApJ/708/661/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        if row['f_SN'] == 'a':
            name = 'SDSS-II ' + str(row['SN'])
        else:
            name = 'SN' + row['SN']
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2010ApJ...708..661D')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity('redshift', str(
            row['z']), source, error=str(row['e_z']))
    catalog.journal_entries()

    # 2014ApJ...795...44R
    result = Vizier.get_catalogs('J/ApJ/795/44/ps1_snIa')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        name = row['SN']
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2014ApJ...795...44R')
        catalog.entries[name].add_quantity('alias', name, source)
        astrot = astrotime(row['tdisc'], format='mjd').datetime
        catalog.entries[name].add_quantity('discoverdate',  make_date_string(
            astrot.year, astrot.month, astrot.day), source)
        catalog.entries[name].add_quantity('redshift', str(
            row['z']), source, error=str(row['e_z']), kind='heliocentric')
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
        catalog.entries[name].add_quantity('claimedtype', 'Ia', source)

    result = Vizier.get_catalogs('J/ApJ/795/44/table6')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        name = row['SN']
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2014ApJ...795...44R')
        catalog.entries[name].add_quantity('alias', name, source)
        if row['mag'] != '--':
            catalog.entries[name].add_photometry(
                time=str(row['MJD']), band=row['Filt'],
                magnitude=str(row['mag']),
                e_magnitude=str(row['e_mag']), source=source, system='AB',
                telescope='PS1', instrument='PS1')
    catalog.journal_entries()

    # 1990A&AS...82..145C
    result = Vizier.get_catalogs('II/189/mag')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)

    with open(os.path.join(
            catalog.get_current_task_repo(), 'II_189_refs.csv')) as f:
        tsvin = csv.reader(f, delimiter='\t', skipinitialspace=True)
        ii189bibdict = {}
        ii189refdict = {}
        for r, row in enumerate(tsvin):
            if row[0] != '0':
                ii189bibdict[r + 1] = row[1]
            else:
                ii189refdict[r + 1] = row[2]

    for row in pbar(table, task_str):
        if row['band'][0] == '(':
            continue
        oldname = 'SN' + row['SN']
        name = catalog.add_entry(oldname)
        source = ''
        secsource = catalog.entries[name].add_source(
            bibcode='1990A&AS...82..145C', secondary=True)
        mjd = str(jd_to_mjd(Decimal(row['JD'])))
        mag = str(row['m'])
        band = row['band'].strip("'")
        if row['r_m'] in ii189bibdict:
            source = catalog.entries[name].add_source(
                bibcode=ii189bibdict[row['r_m']])
        else:
            source = catalog.entries[name].add_source(
                name=ii189refdict[row['r_m']])
        catalog.entries[name].add_quantity('alias', oldname, source)

        catalog.entries[name].add_photometry(
            time=mjd, band=band, magnitude=mag,
            source=uniq_cdl([source, secsource]))
    catalog.journal_entries()

    # 2014yCat.7272....0G
    result = Vizier.get_catalogs('VII/272/snrs')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)

    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = ''
        if row['Names']:
            names = row['Names'].split(',')
            for nam in names:
                if nam.strip()[:2] == 'SN':
                    name = nam.strip()
                    if is_number(name[2:]):
                        name = name + 'A'
            if not name:
                for nam in names:
                    if nam.strip('()') == nam:
                        name = nam.strip()
                        break
        if not name:
            name = row['SNR'].strip()

        oldname = name
        name = catalog.add_entry(oldname)
        source = (catalog.entries[name]
                  .add_source(bibcode='2014BASI...42...47G') +
                  ',' +
                  (catalog.entries[name]
                   .add_source(name='Galactic SNRs',
                               url=('https://www.mrao.cam.ac.uk/'
                                    'surveys/snrs/snrs.data.html'))))
        catalog.entries[name].add_quantity('alias', oldname, source)

        catalog.entries[name].add_quantity('alias', row['SNR'].strip(), source)
        catalog.entries[name].add_quantity(
            'alias', 'MWSNR ' + row['SNR'].strip('G '), source)

        if row['Names']:
            names = row['Names'].split(',')
            for nam in names:
                catalog.entries[name].add_quantity('alias', nam.replace(
                    'Vela (XYZ)', 'Vela').strip('()').strip(), source)
                if nam.strip()[:2] == 'SN':
                    catalog.entries[name].add_quantity(
                        'discoverdate', nam.strip()[2:], source)

        catalog.entries[name].add_quantity('host', 'Milky Way', source)
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
    catalog.journal_entries()

    # 2014MNRAS.442..844F
    result = Vizier.get_catalogs('J/MNRAS/442/844/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = 'SN' + row['SN']
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2014MNRAS.442..844F')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity('redshift', str(
            row['zhost']), source, kind='host')
        catalog.entries[name].add_quantity('ebv', str(row['E_B-V_']), source)
    catalog.journal_entries()

    result = Vizier.get_catalogs('J/MNRAS/442/844/table2')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    instr = 'KAIT'
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = 'SN' + str(row['SN'])
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2014MNRAS.442..844F')
        catalog.entries[name].add_quantity('alias', name, source)
        for band in ['B', 'V', 'R', 'I']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=row['MJD'], band=band, magnitude=row[bandtag],
                    e_magnitude=row['e_' + bandtag], source=source,
                    telescope=instr, instrument=instr)
    catalog.journal_entries()

    # 2012MNRAS.425.1789S
    result = Vizier.get_catalogs('J/MNRAS/425/1789/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = ''.join(row['SimbadName'].split(' '))
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2012MNRAS.425.1789S')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity('alias', 'SN' + row['SN'], source)
        catalog.entries[name].add_quantity('host', row['Gal'], source)
        if is_number(row['cz']):
            red_str = str(
                round_sig(float(row['cz']) * KM / CLIGHT,
                          sig=get_sig_digits(str(row['cz']))))
            catalog.entries[name].add_quantity(
                'redshift', red_str, source, kind='heliocentric')
        catalog.entries[name].add_quantity('ebv', str(row['E_B-V_']), source)
    catalog.journal_entries()

    # 2015ApJS..219...13W
    result = Vizier.get_catalogs('J/ApJS/219/13/table3')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = u'LSQ' + str(row['LSQ'])
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2015ApJS..219...13W')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
        catalog.entries[name].add_quantity(
            'redshift', row['z'], source,
            error=row['e_z'], kind='heliocentric')
        catalog.entries[name].add_quantity('ebv', row['E_B-V_'], source)
        catalog.entries[name].add_quantity('claimedtype', 'Ia', source)
    result = Vizier.get_catalogs('J/ApJS/219/13/table2')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = 'LSQ' + row['LSQ']
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2015ApJS..219...13W')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_photometry(
            time=str(jd_to_mjd(Decimal(row['JD']))),
            instrument='QUEST',
            observatory='La Silla', band=row['Filt'], telescope='ESO Schmidt',
            magnitude=row['mag'], e_magnitude=row['e_mag'], system='Swope',
            source=source)
    catalog.journal_entries()

    # 2012Natur.491..228C
    result = Vizier.get_catalogs('J/other/Nat/491.228/tablef1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    name = 'SN2213-1745'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2012Natur.491..228C')
    catalog.entries[name].add_quantity('alias', name, source)
    catalog.entries[name].add_quantity('claimedtype', 'SLSN-R', source)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['g', 'r', 'i']:
            bandtag = band + '_mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=row['MJD' + band + '_'], band=band + "'",
                    magnitude=row[bandtag], e_magnitude=row['e_' + bandtag],
                    source=source)

    result = Vizier.get_catalogs('J/other/Nat/491.228/tablef2')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    name = 'SN1000+0216'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2012Natur.491..228C')
    catalog.entries[name].add_quantity('alias', name, source)
    catalog.entries[name].add_quantity('claimedtype', 'SLSN-II?', source)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['g', 'r', 'i']:
            bandtag = band + '_mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=row['MJD' + band + '_'], band=band + "'",
                    magnitude=row[bandtag], e_magnitude=row['e_' + bandtag],
                    source=source)
    catalog.journal_entries()

    # 2011Natur.474..484Q
    result = Vizier.get_catalogs('J/other/Nat/474.484/tables1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = str(row['Name'])
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2011Natur.474..484Q')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_photometry(
            time=row['MJD'], band=row['Filt'], telescope=row['Tel'],
            magnitude=row['mag'], e_magnitude=row['e_mag'], source=source)
    catalog.journal_entries()

    # 2011ApJ...736..159G
    result = Vizier.get_catalogs('J/ApJ/736/159/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    name = 'PTF10vdl'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2011ApJ...736..159G')
    catalog.entries[name].add_quantity('alias', name, source)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        catalog.entries[name].add_photometry(
            time=str(jd_to_mjd(Decimal(row['JD']))),
            band=row['Filt'],
            telescope=row['Tel'], magnitude=row['mag'],
            e_magnitude=row['e_mag'] if is_number(row['e_mag']) else '',
            upperlimit=(not is_number(row['e_mag'])), source=source)
    catalog.journal_entries()

    # 2012ApJ...760L..33B
    result = Vizier.get_catalogs('J/ApJ/760/L33/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    name = 'PTF12gzk'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2012ApJ...760L..33B')
    catalog.entries[name].add_quantity('alias', name, source)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        # Fixing a typo in VizieR table
        if str(row['JD']) == '2455151.456':
            row['JD'] = '2456151.456'
        catalog.entries[name].add_photometry(
            time=str(jd_to_mjd(Decimal(row['JD']))),
            band=row['Filt'],
            telescope=row['Inst'], magnitude=row['mag'],
            e_magnitude=row['e_mag'], source=source)
    catalog.journal_entries()

    # 2013ApJ...769...39S
    result = Vizier.get_catalogs('J/ApJ/769/39/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    name = 'PS1-12sk'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2013ApJ...769...39S')
    catalog.entries[name].add_quantity('alias', name, source)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        instrument = ''
        telescope = ''
        if row['Inst'] == 'RATCam':
            instrument = row['Inst']
        else:
            telescope = row['Inst']
        catalog.entries[name].add_photometry(
            time=row['MJD'], band=row['Filt'], telescope=telescope,
            instrument=instrument, magnitude=row['mag'],
            e_magnitude=row['e_mag'] if not row['l_mag'] else '',
            upperlimit=(row['l_mag'] == '>'), source=source)
    catalog.journal_entries()

    # 2009MNRAS.394.2266P
    # Note: Instrument info available via links in VizieR, can't auto-parse
    # just yet.
    name = 'SN2005cs'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2009MNRAS.394.2266P')
    catalog.entries[name].add_quantity('alias', name, source)
    result = Vizier.get_catalogs('J/MNRAS/394/2266/table2')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['U', 'B', 'V', 'R', 'I']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                e_mag = (row['e_' + bandtag]
                         if row['l_' + bandtag] != '>' else '')
                upl = row['l_' + bandtag] == '>'
                catalog.entries[name].add_photometry(
                    time=str(jd_to_mjd(Decimal(row['JD']))),
                    band=band,
                    magnitude=row[bandtag], e_magnitude=e_mag,
                    source=source, upperlimit=upl)
        if ('zmag' in row and is_number(row['zmag']) and not
                isnan(float(row['zmag']))):
            catalog.entries[name].add_photometry(
                time=str(jd_to_mjd(Decimal(row['JD']))),
                band='z',
                magnitude=row['zmag'], e_magnitude=row['e_zmag'],
                source=source)

    result = Vizier.get_catalogs('J/MNRAS/394/2266/table3')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['B', 'V', 'R']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                time = str(jd_to_mjd(Decimal(row['JD'])))
                e_mag = (row['e_' + bandtag]
                         if row['l_' + bandtag] != '>' else '')
                catalog.entries[name].add_photometry(
                    time=time, band=band, magnitude=row[bandtag],
                    e_magnitude=e_mag, source=source,
                    upperlimit=(row['l_' + bandtag] == '>'))

    result = Vizier.get_catalogs('J/MNRAS/394/2266/table4')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['J', 'H', 'K']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=str(jd_to_mjd(Decimal(row['JD']))),
                    band=band, magnitude=row[bandtag],
                    e_magnitude=row['e_' + bandtag], source=source)
    catalog.journal_entries()

    # 2013AJ....145...99A
    result = Vizier.get_catalogs('J/AJ/145/99/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    name = 'SN2003ie'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2013AJ....145...99A')
    catalog.entries[name].add_quantity('alias', name, source)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['B', 'R']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=row["MJD"], band=band,
                    magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag] if
                    not row["l_" + bandtag] else '',
                    upperlimit=(row['l_' + bandtag] == '>'),
                    source=source)
        for band in ['V', 'I']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=row["MJD"], band=band,
                    magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag] if
                    is_number(row["e_" + bandtag]) else '',
                    upperlimit=(not is_number(row["e_" + bandtag])),
                    source=source)
    catalog.journal_entries()

    # 2011ApJ...729..143C
    name = 'SN2008am'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2011ApJ...729..143C')
    catalog.entries[name].add_quantity('alias', name, source)

    result = Vizier.get_catalogs('J/ApJ/729/143/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        e_mag = row['e_mag'] if not row['l_mag'] else ''
        catalog.entries[name].add_photometry(
            time=row['MJD'], band='ROTSE', telescope='ROTSE',
            magnitude=row['mag'], e_magnitude=e_mag,
            upperlimit=(row['l_mag'] == '<'), source=source)

    result = Vizier.get_catalogs('J/ApJ/729/143/table2')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['J', 'H', 'Ks']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=row["MJD"],
                    telescope="PAIRITEL", band=band,
                    magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag], source=source)

    result = Vizier.get_catalogs('J/ApJ/729/143/table4')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        catalog.entries[name].add_photometry(
            time=row['MJD'], band=row['Filt'], telescope='P60',
            magnitude=row['mag'], e_magnitude=row['e_mag'], source=source)

    result = Vizier.get_catalogs('J/ApJ/729/143/table5')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        catalog.entries[name].add_photometry(
            time=row['MJD'], band=row['Filt'], instrument='UVOT',
            telescope='Swift',
            magnitude=row['mag'], e_magnitude=row['e_mag'], source=source)
    catalog.journal_entries()

    # 2011ApJ...728...14P
    name = 'SN2009bb'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2011ApJ...728...14P')
    catalog.entries[name].add_quantity('alias', name, source)

    result = Vizier.get_catalogs('J/ApJ/728/14/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['B', 'V', 'R', 'I']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=str(jd_to_mjd(Decimal(row["JD"]))),
                    telescope=row["Tel"], band=band,
                    magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag], source=source)

    result = Vizier.get_catalogs('J/ApJ/728/14/table2')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['u', 'g', 'r', 'i', 'z']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=str(jd_to_mjd(Decimal(row["JD"]))),
                    telescope=row["Tel"], band=band + "'",
                    magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag], source=source)

    result = Vizier.get_catalogs('J/ApJ/728/14/table3')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['Y', 'J', 'H']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=str(jd_to_mjd(Decimal(row["JD"]))),
                    instrument=row['Inst'], band=band,
                    magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag], source=source)
    catalog.journal_entries()

    # 2011PAZh...37..837T
    name = 'SN2009nr'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2011PAZh...37..837T')
    catalog.entries[name].add_quantity('alias', name, source)

    result = Vizier.get_catalogs('J/PAZh/37/837/table2')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        mjd = str(jd_to_mjd(Decimal(row['JD']) + 2455000))
        for band in ['U', 'B', 'V', 'R', 'I']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=mjd, telescope=row["Tel"], band=band,
                    magnitude=row[bandtag], e_magnitude=row["e_" + bandtag],
                    source=source)
    catalog.journal_entries()

    # 2013MNRAS.433.1871B
    name = 'SN2012aw'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2013MNRAS.433.1871B')
    catalog.entries[name].add_quantity('alias', name, source)

    result = Vizier.get_catalogs('J/MNRAS/433/1871/table3a')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        mjd = str(jd_to_mjd(Decimal(row['JD']) + 2456000))
        for band in ['U', 'B', 'V', 'Rc', 'Ic']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=mjd, telescope=row["Tel"],
                    band=band, magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag], source=source)

    result = Vizier.get_catalogs('J/MNRAS/433/1871/table3b')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        mjd = str(jd_to_mjd(Decimal(row['JD']) + 2456000))
        for band in ['g', 'r', 'i', 'z']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=mjd, telescope=row["Tel"],
                    band=band, magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag], source=source)
    catalog.journal_entries()

    # 2014AJ....148....1Z
    name = 'SN2012fr'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2014AJ....148....1Z')
    catalog.entries[name].add_quantity('alias', name, source)

    result = Vizier.get_catalogs('J/AJ/148/1/table2')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        mjd = row['MJD']
        for band in ['B', 'V', 'R', 'I']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=mjd, telescope="LJT",
                    instrument="YFOSC", band=band,
                    magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag], source=source)

    result = Vizier.get_catalogs('J/AJ/148/1/table3')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        mjd = row['MJD']
        for band in ['U', 'B', 'V', 'UVW1', 'UVW2', 'UVM2']:
            bandtag = band + 'mag' if len(band) == 1 else band
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=mjd, telescope="Swift",
                    instrument="UVOT", band=band,
                    magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag], source=source)

    result = Vizier.get_catalogs('J/AJ/148/1/table5')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        mjd = row['MJD']
        for band in ['B', 'V', 'R', 'I']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=mjd, telescope="LJT",
                    band=band, magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag], source=source)
    catalog.journal_entries()

    # 2015ApJ...805...74B
    name = 'SN2014J'
    name = catalog.add_entry(name)
    source = catalog.entries[name].add_source(bibcode='2014AJ....148....1Z')
    catalog.entries[name].add_quantity('alias', name, source)

    result = Vizier.get_catalogs('J/ApJ/805/74/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        mjd = row['MJD']
        if ('mag' in row and is_number(row['mag']) and not
                isnan(float(row['mag']))):
            catalog.entries[name].add_photometry(
                time=mjd, telescope='Swift',
                instrument='UVOT', band=row['Filt'],
                magnitude=row['mag'],
                e_magnitude=row['e_mag'], source=source)
        elif ('maglim' in row and is_number(row['maglim']) and not
              isnan(float(row['maglim']))):
            catalog.entries[name].add_photometry(
                time=mjd, telescope='Swift',
                instrument='UVOT', band=row['Filt'],
                magnitude=row['maglim'],
                upperlimit=True, source=source)
    catalog.journal_entries()

    # 2011ApJ...741...97D
    result = Vizier.get_catalogs('J/ApJ/741/97/table2')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = str(row['SN'])
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2011ApJ...741...97D')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_photometry(
            time=str(jd_to_mjd(Decimal(row['JD']))),
            band=row['Filt'], magnitude=row['mag'],
            e_magnitude=row['e_mag'] if
            is_number(row['e_mag']) else '',
            upperlimit=(not is_number(row['e_mag'])), source=source)
    catalog.journal_entries()

    # 2015MNRAS.448.1206M
    # Note: Photometry from two SN can also be added from this source.
    result = Vizier.get_catalogs('J/MNRAS/448/1206/table3')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = str(row['Name'])
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2015MNRAS.448.1206M')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity(
            'discoverdate', '20' + name[4:6], source)
        catalog.entries[name].add_quantity(
            'ra', row['RAJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['DEJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'redshift', row['zsp'], source, kind='spectroscopic')
        catalog.entries[name].add_quantity(
            'maxappmag', row['rP1mag'], source, error=row['e_rP1mag'])
        catalog.entries[name].add_quantity('maxband', 'r', source)
        catalog.entries[name].add_quantity('claimedtype', 'Ia', source)
    result = Vizier.get_catalogs('J/MNRAS/448/1206/table4')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = str(row['Name'])
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2015MNRAS.448.1206M')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity(
            'discoverdate', '20' + name[4:6], source)
        catalog.entries[name].add_quantity(
            'ra', row['RAJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['DEJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'redshift', row['zph'], source, error=row['e_zph'],
            kind='photometric')
        catalog.entries[name].add_quantity(
            'maxappmag', row['rP1mag'], source, error=row['e_rP1mag'])
        catalog.entries[name].add_quantity('maxband', 'r', source)
        catalog.entries[name].add_quantity('claimedtype', 'Ia?', source)
    result = Vizier.get_catalogs('J/MNRAS/448/1206/table5')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = str(row['Name'])
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2015MNRAS.448.1206M')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity(
            'discoverdate', '20' + name[4:6], source)
        catalog.entries[name].add_quantity(
            'ra', row['RAJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['DEJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'redshift', row['zsp'], source, kind='spectroscopic')
        catalog.entries[name].add_quantity(
            'maxappmag', row['rP1mag'], source, error=row['e_rP1mag'])
        catalog.entries[name].add_quantity('maxband', 'r', source)
        catalog.entries[name].add_quantity('claimedtype', row['Type'], source)
    result = Vizier.get_catalogs('J/MNRAS/448/1206/table6')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = str(row['Name'])
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2015MNRAS.448.1206M')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity(
            'discoverdate', '20' + name[4:6], source)
        catalog.entries[name].add_quantity(
            'ra', row['RAJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['DEJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'maxappmag', row['rP1mag'], source, error=row['e_rP1mag'])
        catalog.entries[name].add_quantity('maxband', 'r', source)
        catalog.entries[name].add_quantity('claimedtype', row['Type'], source)
    result = Vizier.get_catalogs('J/MNRAS/448/1206/tablea2')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = str(row['Name'])
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2015MNRAS.448.1206M')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity(
            'discoverdate', '20' + name[4:6], source)
        catalog.entries[name].add_quantity(
            'ra', row['RAJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['DEJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'maxappmag', row['rP1mag'], source, error=row['e_rP1mag'])
        catalog.entries[name].add_quantity('maxband', 'r', source)
        catalog.entries[name].add_quantity(
            'claimedtype', row['Typesoft'] + '?', source)
        catalog.entries[name].add_quantity(
            'claimedtype', row['Typepsnid'] + '?', source)
    result = Vizier.get_catalogs('J/MNRAS/448/1206/tablea3')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = str(row['Name'])
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2015MNRAS.448.1206M')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity(
            'discoverdate', '20' + name[4:6], source)
        catalog.entries[name].add_quantity(
            'ra', row['RAJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['DEJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'maxappmag', row['rP1mag'], source, error=row['e_rP1mag'])
        catalog.entries[name].add_quantity('maxband', 'r', source)
        catalog.entries[name].add_quantity('claimedtype', 'Candidate', source)
    catalog.journal_entries()

    # 2012AJ....143..126B
    result = Vizier.get_catalogs('J/AJ/143/126/table4')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        if not row['Wcl'] or row['Wcl'] == 'N':
            continue
        row = convert_aq_output(row)
        name = str(row['SN']).replace(' ', '')
        name = catalog.add_entry(name)
        source = catalog.entries[name].add_source(
            bibcode='2012AJ....143..126B')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity(
            'claimedtype', 'Ia-' + row['Wcl'], source)
    catalog.journal_entries()

    # 2015ApJS..220....9F
    for viztab in ['1', '2']:
        result = Vizier.get_catalogs('J/ApJS/220/9/table' + viztab)
        table = result[list(result.keys())[0]]
        table.convert_bytestring_to_unicode(python3_only=True)
        for row in pbar(table, task_str):
            row = convert_aq_output(row)
            name = catalog.add_entry(name=row['SN'])
            source = catalog.entries[name].add_source(
                bibcode='2015ApJS..220....9F')
            catalog.entries[name].add_quantity('alias', name, source)
            catalog.entries[name].add_quantity(
                'claimedtype', row['Type'], source)
            catalog.entries[name].add_quantity(
                'ra', row['RAJ2000'], source, unit='floatdegrees')
            catalog.entries[name].add_quantity(
                'dec', row['DEJ2000'], source, unit='floatdegrees')
            if '?' not in row['Host']:
                catalog.entries[name].add_quantity(
                    'host', row['Host'].replace('_', ' '), source)
            kind = ''
            if 'Host' in row['n_z']:
                kind = 'host'
            elif 'Spectrum' in row['n_z']:
                kind = 'spectroscopic'
            catalog.entries[name].add_quantity(
                'redshift', row['z'], source, error=row['e_z'], kind=kind)

    result = Vizier.get_catalogs('J/ApJS/220/9/table8')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = catalog.add_entry(row['SN'])
        source = catalog.entries[name].add_source(
            bibcode='2015ApJS..220....9F')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity('claimedtype', row['Type'], source)
        catalog.entries[name].add_photometry(
            time=row['MJD'], band=row['Band'], magnitude=row['mag'],
            e_magnitude=row['e_mag'], telescope=row['Tel'], source=source)
    catalog.journal_entries()

    result = Vizier.get_catalogs('J/ApJ/673/999/table1')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = catalog.add_entry(name='SN' + row['SN'])
        source = catalog.entries[name].add_source(
            bibcode='2008ApJ...673..999P')
        catalog.entries[name].add_quantity('alias', name, source)
        catalog.entries[name].add_quantity(
            'ra', row['RAJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['DEJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'redshift', row['z'], source, kind='host')
        catalog.entries[name].add_quantity(
            'hostra', row['RAGdeg'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'hostdec', row['DEGdeg'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'claimedtype', row['Type'].strip(':'), source)
    catalog.journal_entries()

    # 2011MNRAS.417..916G
    result = Vizier.get_catalogs("J/MNRAS/417/916/table2")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name, source = catalog.new_entry(
            'SNSDF' + row['SNSDF'],
            bibcode="2011MNRAS.417..916G")
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
        catalog.entries[name].add_quantity(
            'redshift', row['zsp'] if row['zsp'] else row['zph'], source,
            kind='host')
        catalog.entries[name].add_quantity(
            'discoverdate', '20' + row['SNSDF'][:2] + '/' + row['SNSDF'][2:4],
            source, kind='host')
        catalog.entries[name].add_quantity(
            'hostoffsetang', row['Offset'], source, unit='arcseconds')
        catalog.entries[name].add_quantity('claimedtype', row['Type'], source)
    catalog.journal_entries()

    # 2013MNRAS.430.1746G
    result = Vizier.get_catalogs("J/MNRAS/430/1746/table4")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name, source = catalog.new_entry(
            'SDSS' + row['SDSS'],
            bibcode="2013MNRAS.430.1746G")
        catalog.entries[name].add_quantity(
            'ra', row['RAJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['DEJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'discoverdate', row['Date'].replace('-', '/'), source)
        catalog.entries[name].add_quantity('redshift', row['z'], source)
        catalog.entries[name].add_quantity('claimedtype', row['Type'], source)
    catalog.journal_entries()

    # 2014AJ....148...13R
    result = Vizier.get_catalogs("J/AJ/148/13/high_z")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name, source = catalog.new_entry(
            row['Name'],
            bibcode="2014AJ....148...13R")
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
        catalog.entries[name].add_quantity(
            'discoverdate', '20' + row['Name'][3:5], source)
        catalog.entries[name].add_quantity(
            'redshift', row['zSN'], source, kind='heliocentric',
            error=row['e_zSN'])
        catalog.entries[name].add_quantity('hostra', row['RAG'], source)
        catalog.entries[name].add_quantity('hostdec', row['DEG'], source)
        catalog.entries[name].add_quantity(
            'hostoffsetang', row['ASep'], source, unit='arcseconds')
        catalog.entries[name].add_quantity(
            'redshift', row['zhost'], source, kind='host',
            error=row['e_zhost'])
    result = Vizier.get_catalogs("J/AJ/148/13/low_z")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name, source = catalog.new_entry(
            row['Name'],
            bibcode="2014AJ....148...13R")
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
        catalog.entries[name].add_quantity(
            'discoverdate', '20' + row['Name'][3:5], source)
        catalog.entries[name].add_quantity(
            'redshift', row['zSN'], source, kind='heliocentric',
            error=row['e_zSN'])
        catalog.entries[name].add_quantity('hostra', row['RAG'], source)
        catalog.entries[name].add_quantity('hostdec', row['DEG'], source)
        catalog.entries[name].add_quantity(
            'hostoffsetang', row['ASep'], source, unit='arcseconds')
        catalog.entries[name].add_quantity(
            'redshift', row['zhost'], source, kind='host',
            error=row['e_zhost'])
    catalog.journal_entries()

    # 2007ApJ...666..674M
    result = Vizier.get_catalogs("J/ApJ/666/674/table3")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        essname = 'ESSENCE ' + row['ESSENCE']
        if row['SN']:
            name = 'SN' + row['SN']
        else:
            name = essname
        name, source = catalog.new_entry(
            name, bibcode="2007ApJ...666..674M")
        catalog.entries[name].add_quantity('alias', essname, source)
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
        catalog.entries[name].add_quantity(
            'redshift', row['zSN'], source, error=row['e_zSN'],
            kind='heliocentric')
        catalog.entries[name].add_quantity(
            'redshift', row['zGal'], source, kind='host')
        catalog.entries[name].add_quantity(
            'claimedtype', row['SType'] if row['SType'] else row['Type'],
            source)
    catalog.journal_entries()

    # 2013AcA....63....1K
    result = Vizier.get_catalogs("J/AcA/63/1/table1")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        if 'OGLE' not in row['Name']:
            continue
        name, source = catalog.new_entry(
            row['Name'],
            bibcode="2013AcA....63....1K")
        catalog.entries[name].add_quantity('alias', row['OGLEIV'], source)
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
        astrot = astrotime(float(row['Tmax']), format='jd').datetime
        catalog.entries[name].add_quantity('maxdate', make_date_string(
            astrot.year, astrot.month, astrot.day), source)
    catalog.journal_entries()

    # 2011MNRAS.410.1262W
    result = Vizier.get_catalogs("J/MNRAS/410/1262/tablea2")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name, source = catalog.new_entry(
            'SNLS-' + row['SN'],
            bibcode="2011MNRAS.410.1262W")
        catalog.entries[name].add_quantity(
            'ra', row['_RA'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['_DE'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'redshift', row['z'], source, error=row['e_z'],
            kind='heliocentric')
    catalog.journal_entries()

    # 2012ApJ...755...61S
    result = Vizier.get_catalogs("J/ApJ/755/61/table3")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        sdssname = 'SDSS-II SN ' + row['SNID']
        if row['SN']:
            name = 'SN' + row['SN']
        else:
            name = sdssname
        name, source = catalog.new_entry(
            name, bibcode="2012ApJ...755...61S")
        err = row['e_z'] if is_number(row['e_z']) else ''
        catalog.entries[name].add_quantity('alias', sdssname, source)
        catalog.entries[name].add_quantity('hostra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('hostdec', row['DEJ2000'], source)
        catalog.entries[name].add_quantity(
            'redshift', row['z'], source, error=err, kind='host')
    catalog.journal_entries()

    # 2008AJ....135..348S
    result = Vizier.get_catalogs("J/AJ/135/348/SNe")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        sdssname = 'SDSS-II SN ' + row['SNID']
        if row['SN']:
            name = 'SN' + row['SN']
        else:
            name = sdssname
        name, source = catalog.new_entry(
            name, bibcode="2008AJ....135..348S")
        catalog.entries[name].add_quantity('alias', sdssname, source)
        fra = Decimal(row['RAJ2000'])
        if fra < Decimal(0.0):
            fra = Decimal(360.0) + fra
        catalog.entries[name].add_quantity(
            'ra', str(fra), source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['DEJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'redshift', row['zsp'], source, kind='spectroscopic')
        catalog.entries[name].add_quantity(
            'claimedtype', row['Type'].replace('SN', '').strip(), source)
    catalog.journal_entries()

    # 2010ApJ...713.1026D
    result = Vizier.get_catalogs("J/ApJ/713/1026/SNe")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        sdssname = 'SDSS-II SN ' + row['ID']
        if row['IAU']:
            name = 'SN' + row['IAU']
        else:
            name = sdssname
        name, source = catalog.new_entry(
            name, bibcode="2010ApJ...713.1026D")
        catalog.entries[name].add_quantity('alias', sdssname, source)
        catalog.entries[name].add_quantity(
            'ra', row['RAJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['DEJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'redshift', row['z'], source, kind='heliocentric')
    catalog.journal_entries()

    # 2013ApJ...770..107C
    result = Vizier.get_catalogs("J/ApJ/770/107/galaxies")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name, source = catalog.new_entry(
            row['SN'], bibcode="2013ApJ...770..107C")
        catalog.entries[name].add_quantity('hostra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('hostdec', row['DEJ2000'], source)
        catalog.entries[name].add_quantity(
            'redshift', row['z'], source,
            error=row['e_z'] if is_number(row['e_z']) else '', kind='host')
    catalog.journal_entries()

    # 2011ApJ...738..162S
    result = Vizier.get_catalogs("J/ApJ/738/162/table3")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = 'SDSS-II SN ' + row['CID']
        name, source = catalog.new_entry(
            name, bibcode="2011ApJ...738..162S")
        fra = Decimal(row['RAJ2000'])
        if fra < Decimal(0.0):
            fra = Decimal(360.0) + fra
        catalog.entries[name].add_quantity(
            'ra', str(fra), source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['DEJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'redshift', row['z'], source, kind='spectroscopic',
            error=row['e_z'])
        catalog.entries[name].add_quantity(
            'claimedtype', 'Ia', source, probability=row['PzIa'])
    result = Vizier.get_catalogs("J/ApJ/738/162/table4")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = 'SDSS-II SN ' + row['CID']
        name, source = catalog.new_entry(
            name, bibcode="2011ApJ...738..162S")
        fra = Decimal(row['RAJ2000'])
        if fra < Decimal(0.0):
            fra = Decimal(360.0) + fra
        catalog.entries[name].add_quantity(
            'ra', str(fra), source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'dec', row['DEJ2000'], source, unit='floatdegrees')
        catalog.entries[name].add_quantity(
            'redshift', row['zph'], source, kind='photometric')
        catalog.entries[name].add_quantity(
            'claimedtype', 'Ia', source, probability=row['PIa'])
    catalog.journal_entries()

    # 2015MNRAS.446..943V
    snrtabs = ["ngc2403", "ngc2903", "ngc300", "ngc3077", "ngc4214", "ngc4395",
               "ngc4449", "ngc5204", "ngc5585", "ngc6946", "ngc7793", "m33",
               "m74", "m81", "m82", "m83", "m101", "m31"]
    for tab in pbar(snrtabs, task_str):
        result = Vizier.get_catalogs("J/MNRAS/446/943/" + tab)
        table = result[list(result.keys())[0]]
        table.convert_bytestring_to_unicode(python3_only=True)
        for ri, row in enumerate(pbar(table, task_str)):
            ra = (row['RAJ2000'] if isinstance(row['RAJ2000'], str) else
                  radec_clean(str(row['RAJ2000']), 'ra',
                              unit='floatdegrees')[0])
            dec = (row['DEJ2000'] if isinstance(row['DEJ2000'], str) else
                   radec_clean(str(row['DEJ2000']), 'dec',
                               unit='floatdegrees')[0])
            name = (tab.upper() + 'SNR J' + rep_chars(ra, ' :.') +
                    rep_chars(dec, ' :.'))
            name, source = catalog.new_entry(
                name, bibcode="2015MNRAS.446..943V")
            catalog.entries[name].add_quantity('ra', ra, source)
            catalog.entries[name].add_quantity('dec', dec, source)
            catalog.entries[name].add_quantity('host', tab.upper(), source)
    catalog.journal_entries()

    # 2009ApJ...703..370C
    result = Vizier.get_catalogs("J/ApJ/703/370/tables")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        ra = row['RAJ2000']
        dec = row['DEJ2000']
        name = row['Gal'].replace(' ', '') + 'SNR J' + \
            rep_chars(ra, ' .') + rep_chars(dec, ' .')
        name, source = catalog.new_entry(
            name, bibcode="2009ApJ...703..370C")
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
        catalog.entries[name].add_quantity('host', row['Gal'], source)
    catalog.journal_entries()

    # 2016ApJ...821...57D
    name, source = catalog.new_entry(
        'SN2013ge', bibcode="2016ApJ...821...57D")
    result = Vizier.get_catalogs("J/ApJ/821/57/table1")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['UVW2', 'UVM2', 'UVW1', 'U', 'B', 'V']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=str(row["MJD"]), band=band,
                    magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag],
                    telescope='Swift', instrument='UVOT',
                    source=source)
    result = Vizier.get_catalogs("J/ApJ/821/57/table2")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['B', 'V', 'R', 'I']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=str(row["MJD"]), band=band,
                    magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag],
                    instrument='CAO', source=source)
    result = Vizier.get_catalogs("J/ApJ/821/57/table3")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['B', 'V', "r'", "i'"]:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                catalog.entries[name].add_photometry(
                    time=str(row["MJD"]), band=band,
                    magnitude=row[bandtag],
                    e_magnitude=row["e_" + bandtag],
                    instrument='FLWO', source=source)
    result = Vizier.get_catalogs("J/ApJ/821/57/table4")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        for band in ['r', 'i', 'z']:
            bandtag = band + 'mag'
            if (bandtag in row and is_number(row[bandtag]) and not
                    isnan(float(row[bandtag]))):
                upp = False
                if "l_" + bandtag in row and row["l_" + bandtag] == ">":
                    upp = True
                catalog.entries[name].add_photometry(
                    time=str(row["MJD"]), band=band,
                    magnitude=row[bandtag], upperlimit=upp,
                    e_magnitude=row["e_" + bandtag] if
                    is_number(row["e_" + bandtag]) else '',
                    instrument=row["Inst"], source=source)
    catalog.journal_entries()

    # 2004ApJ...607..665R
    result = Vizier.get_catalogs("J/ApJ/607/665/table1")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = row['Name'].replace('SN ', 'SN')
        name, source = catalog.new_entry(
            name, bibcode="2004ApJ...607..665R")
        catalog.entries[name].add_quantity('alias', row['OName'], source)
        catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
    result = Vizier.get_catalogs("J/ApJ/607/665/table2")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = row['Name'].replace('SN ', 'SN')
        name, source = catalog.new_entry(
            name, bibcode="2004ApJ...607..665R")
        mjd = str(jd_to_mjd(Decimal(row['HJD'])))
        catalog.entries[name].add_photometry(
            time=mjd, band=row['Filt'],
            magnitude=row['Vega'], system='Vega',
            e_magnitude=row['e_Vega'], source=source)
    result = Vizier.get_catalogs("J/ApJ/607/665/table5")
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = row['Name'].replace('SN ', 'SN')
        name, source = catalog.new_entry(
            name, bibcode="2004ApJ...607..665R")
        catalog.entries[name].add_quantity(
            'redshift', row['z'], source, kind='spectroscopic')
    catalog.journal_entries()

    return


def do_lennarz(catalog):
    """
    """
    task_str = catalog.get_current_task_str()
    Vizier.ROW_LIMIT = -1
    result = Vizier.get_catalogs('J/A+A/538/A120/usc')
    table = result[list(result.keys())[0]]
    table.convert_bytestring_to_unicode(python3_only=True)

    bibcode = '2012A&A...538A.120L'
    for row in pbar(table, task_str):
        row = convert_aq_output(row)
        name = 'SN' + row['SN']
        name = catalog.add_entry(name)

        source = catalog.entries[name].add_source(bibcode=bibcode)
        catalog.entries[name].add_quantity('alias', name, source)

        if row['RAJ2000']:
            catalog.entries[name].add_quantity('ra', row['RAJ2000'], source)
        if row['DEJ2000']:
            catalog.entries[name].add_quantity('dec', row['DEJ2000'], source)
        if row['RAG']:
            catalog.entries[name].add_quantity('hostra', row['RAG'], source)
        if row['DEG']:
            catalog.entries[name].add_quantity('hostdec', row['DEG'], source)
        if row['Gal']:
            catalog.entries[name].add_quantity('host', row['Gal'], source)
        if row['Type']:
            claimedtypes = row['Type'].split('|')
            for claimedtype in claimedtypes:
                catalog.entries[name].add_quantity(
                    'claimedtype', claimedtype.strip(' -'), source)
        if row['z']:
            if name not in ['SN1985D', 'SN2004cq']:
                catalog.entries[name].add_quantity(
                    'redshift', row['z'], source, kind='host')
        if row['Dist']:
            if row['e_Dist']:
                catalog.entries[name].add_quantity(
                    'lumdist', row['Dist'], source,
                    error=row['e_Dist'], kind='host')
            else:
                catalog.entries[name].add_quantity(
                    'lumdist', row['Dist'], source, kind='host')

        if row['Ddate']:
            datestring = row['Ddate'].replace('-', '/')

            catalog.entries[name].add_quantity(
                'discoverdate', datestring, source)

            if 'photometry' not in catalog.entries[name]:
                if ('Dmag' in row and is_number(row['Dmag']) and not
                        isnan(float(row['Dmag']))):
                    datesplit = row['Ddate'].strip().split('-')
                    if len(datesplit) == 3:
                        datestr = row['Ddate'].strip()
                    elif len(datesplit) == 2:
                        datestr = row['Ddate'].strip() + '-01'
                    elif len(datesplit) == 1:
                        datestr = row['Ddate'].strip() + '-01-01'
                    mjd = str(astrotime(datestr).mjd)
                    catalog.entries[name].add_photometry(
                        time=mjd, band=row['Dband'], magnitude=row['Dmag'],
                        source=source)
        if row['Mdate']:
            datestring = row['Mdate'].replace('-', '/')

            catalog.entries[name].add_quantity('maxdate', datestring, source)

            if 'photometry' not in catalog.entries[name]:
                if ('MMag' in row and is_number(row['MMag']) and not
                        isnan(float(row['MMag']))):
                    datesplit = row['Mdate'].strip().split('-')
                    if len(datesplit) == 3:
                        datestr = row['Mdate'].strip()
                    elif len(datesplit) == 2:
                        datestr = row['Mdate'].strip() + '-01'
                    elif len(datesplit) == 1:
                        datestr = row['Mdate'].strip() + '-01-01'
                    mjd = str(astrotime(datestr).mjd)
                    catalog.entries[name].add_photometry(
                        time=mjd, band=row['Mband'],
                        magnitude=row['Mmag'], source=source)

    catalog.journal_entries()
    return



"""
"""
from ..constants import PREF_KINDS

__all__ = ['frame_priority']


def frame_priority(attr):
    if 'kind' in attr:
        if attr['kind'] in PREF_KINDS:
            return PREF_KINDS.index(attr['kind'])
        else:
            return len(PREF_KINDS)
    return len(PREF_KINDS)



"""
"""


def main():
    from datetime import datetime

    beg_time = datetime.now()
    # Process command-line arguments to determine action
    #    If no subcommand (e.g. 'impoter') is given, returns 'None' --> exit
    args = load_args()
    if args is None:
        return

    # FIX
    # LOAD SUPERNOVAE SPECIFIC STUFF EXPLICITLY FOR NOW.  LATER, CHOOSE BASED
    #    ON ARGS WHAT TO IMPORT AND INITIALIZE

    from astrocats.supernovae.catalog import Catalog
    catalog = Catalog(args)
    git_vers = get_git()
    title_str = "Open Supernova Catalog, version: {}".format(git_vers)
    catalog.log.warning("\n\n{}\n{}\n{}\n".format(
        title_str, '=' * len(title_str), beg_time.ctime()))

    # Choose which submodule to run (note: can also use `set_default` with
    # function)
    if args._name == 'sn-import':
        catalog.log.info("Running `importer`.")
        catalog.import_data()

    end_time = datetime.now()
    catalog.log.warning("All complete at {}, After {}".format(
        end_time, end_time - beg_time))
    return


def load_args(args=None):
    import argparse

    parser = argparse.ArgumentParser(
        description='Generate a catalog JSON file and plot HTML files from '
        'SNE data.')
    # parser.add_argument('--foo', action='store_true', help='foo help')
    subparsers = parser.add_subparsers(
        description='valid subcommands', dest='_name',
        help='sub-command help')

    # Build a 'parent' parser whose settings are inhereted by children parsers
    pars_parent = argparse.ArgumentParser(add_help=False)
    pars_parent.add_argument(
        '--verbose', '-v', dest='verbose', default=False, action='store_true',
        help='Print more messages to the screen.')
    pars_parent.add_argument(
        '--debug', '-d', dest='debug', default=False, action='store_true',
        help='Print excessive messages to the screen.')
    pars_parent.add_argument(
        '--travis', '-t',  dest='travis',  default=False, action='store_true',
        help='Run import script in test mode for Travis.')
    pars_parent.add_argument(
        '--log',  dest='log_filename',  default=None,
        help='Filename to which to store logging information.')

    # If output files should be written or not
    # ----------------------------------------
    write_group = pars_parent.add_mutually_exclusive_group()
    write_group.add_argument(
        '--write', action='store_true', dest='write_entries', default=True,
        help='Write entries to files [default].')
    write_group.add_argument(
        '--no-write', action='store_false', dest='write_entries', default=True,
        help='do not write entries to file.')

    # If previously ceared output files should be deleted or not
    # ----------------------------------------------------------
    delete_group = pars_parent.add_mutually_exclusive_group()
    delete_group.add_argument(
        '--predelete', action='store_true', dest='delete_old', default=True,
        help='Delete all old event files to begin [default].')
    delete_group.add_argument(
        '--no-predelete', action='store_false', dest='delete_old',
        default=True, help='Do not delete all old event files to start.')

    # `importer` submodule --- importing supernova data
    pars_imp = subparsers.add_parser("sn-import", parents=[pars_parent],
                                     help="Generate a catalog JSON file")
    pars_imp.add_argument('--update', '-u', dest='update',
                          default=False, action='store_true',
                          help='Only update catalog using live sources.')
    pars_imp.add_argument('--refresh', '-r', dest='refresh',
                          default=False, action='store_true',
                          help='Ignore most task caches.')
    pars_imp.add_argument('--full-refresh', '-f', dest='full_refresh',
                          default=False, action='store_true',
                          help='Ignore all task caches.')
    pars_imp.add_argument('--archived', '-a', dest='archived',
                          default=False, action='store_true',
                          help='Always use task caches.')
    pars_imp.add_argument(
        '--refresh-list', '-rl', dest='refresh_list', default='', nargs='+',
        help='Space-delimited list of caches to clear.')

    # Control which 'tasks' are executed
    # ----------------------------------
    pars_imp.add_argument(
        '--tasks', dest='args_task_list', nargs='*', default=None,
        help='space delimited list of tasks to perform (others disabled).')
    pars_imp.add_argument(
        '--yes', dest='yes_task_list', nargs='+', default=None,
        help='space delimited list of tasks to turn on.')
    pars_imp.add_argument(
        '--no', dest='no_task_list', nargs='+', default=None,
        help='space delimited list of tasks to turn off.')
    pars_imp.add_argument(
        '--min-task-priority', dest='min_task_priority',
        default=None,
        help='minimum priority for a task to run')
    pars_imp.add_argument(
        '--max-task-priority', dest='max_task_priority',
        default=None,
        help='maximum priority for a task to run')
    pars_imp.add_argument(
        '--task-groups', dest='task_groups',
        default=None,
        help='predefined group(s) of tasks to run.')

    args = parser.parse_args(args=args)
    # Print the help information if no subcommand is given
    # subcommand is required for operation
    if args._name is None:
        parser.print_help()
        return None

    return args


def get_git():
    """Get a string representing the current git status --- i.e. tag and commit
    hash.
    """
    import subprocess
    git_vers = subprocess.getoutput(["git describe --always"]).strip()
    return git_vers



"""Imported tasks for the Carnegie Supernova Program.
"""
import csv
import os
from glob import glob

from astrocats.catalog.utils import jd_to_mjd, pbar_strings
from astrocats.supernovae.utils import clean_snname
from cdecimal import Decimal


def do_csp_photo(catalog):
    import re
    cspbands = ['u', 'B', 'V', 'g', 'r', 'i', 'Y', 'J', 'H', 'K']
    file_names = glob(
        os.path.join(catalog.get_current_task_repo(), 'CSP/*.dat'))
    task_str = catalog.get_current_task_str()
    for fname in pbar_strings(file_names, task_str):
        tsvin = csv.reader(open(fname, 'r'), delimiter='\t',
                           skipinitialspace=True)
        eventname = os.path.basename(os.path.splitext(fname)[0])
        eventparts = eventname.split('opt+')
        name = clean_snname(eventparts[0])
        name = catalog.add_entry(name)

        reference = 'Carnegie Supernova Project'
        refbib = '2010AJ....139..519C'
        refurl = 'http://csp.obs.carnegiescience.edu/data'
        source = catalog.entries[name].add_source(
            bibcode=refbib, name=reference, url=refurl)
        catalog.entries[name].add_quantity('alias', name, source)

        year = re.findall(r'\d+', name)[0]
        catalog.entries[name].add_quantity('discoverdate', year, source)

        for r, row in enumerate(tsvin):
            if len(row) > 0 and row[0][0] == "#":
                if r == 2:
                    redz = row[0].split(' ')[-1]
                    catalog.entries[name].add_quantity(
                        'redshift', redz, source, kind='cmb')
                    catalog.entries[name].add_quantity(
                        'ra', row[1].split(' ')[-1], source)
                    catalog.entries[name].add_quantity(
                        'dec', row[2].split(' ')[-1], source)
                continue
            for v, val in enumerate(row):
                if v == 0:
                    mjd = val
                elif v % 2 != 0:
                    if float(row[v]) < 90.0:
                        catalog.entries[name].add_photometry(
                            time=mjd, observatory='LCO',
                            band=cspbands[(v - 1) // 2],
                            system='CSP', magnitude=row[v],
                            e_magnitude=row[v + 1], source=source)

    catalog.journal_entries()
    return


def do_csp_spectra(catalog):
    oldname = ''
    task_str = catalog.get_current_task_str()
    file_names = glob(os.path.join(catalog.get_current_task_repo(), 'CSP/*'))
    for fi, fname in enumerate(pbar_strings(file_names, task_str)):
        filename = os.path.basename(fname)
        sfile = filename.split('.')
        if sfile[1] == 'txt':
            continue
        sfile = sfile[0]
        fileparts = sfile.split('_')
        name = 'SN20' + fileparts[0][2:]
        name = catalog.get_preferred_name(name)
        if oldname and name != oldname:
            catalog.journal_entries()
        oldname = name
        name = catalog.add_entry(name)
        telescope = fileparts[-2]
        instrument = fileparts[-1]
        source = catalog.entries[name].add_source(
            bibcode='2013ApJ...773...53F')
        catalog.entries[name].add_quantity('alias', name, source)

        data = csv.reader(open(fname, 'r'), delimiter=' ',
                          skipinitialspace=True)
        specdata = []
        for r, row in enumerate(data):
            if row[0] == '#JDate_of_observation:':
                jd = row[1].strip()
                time = str(jd_to_mjd(Decimal(jd)))
            elif row[0] == '#Redshift:':
                catalog.entries[name].add_quantity('redshift', row[1].strip(),
                                                   source)
            if r < 7:
                continue
            specdata.append(list(filter(None, [x.strip(' ') for x in row])))
        specdata = [list(i) for i in zip(*specdata)]
        wavelengths = specdata[0]
        fluxes = specdata[1]

        catalog.entries[name].add_spectrum(
            'Angstrom', 'erg/s/cm^2/Angstrom', u_time='MJD',
            time=time, wavelengths=wavelengths, fluxes=fluxes,
            telescope=telescope, instrument=instrument,
            source=source, deredshifted=True, filename=filename)
        if catalog.args.travis and fi >= catalog.TRAVIS_QUERY_LIMIT:
            break

    catalog.journal_entries()
    return



# -*- coding: utf-8 -*-
"""General data import tasks.
"""
import os
from collections import OrderedDict
from glob import glob

from astrocats.catalog.utils import pbar_strings
from astrocats.supernovae.supernova import KEYS, Supernova


def do_external_radio(catalog):
    task_str = catalog.get_current_task_str()
    path_pattern = os.path.join(catalog.get_current_task_repo(), '*.txt')
    for datafile in pbar_strings(glob(path_pattern), task_str):
        oldname = os.path.basename(datafile).split('.')[0]
        name = catalog.add_entry(oldname)
        radiosourcedict = OrderedDict()
        with open(datafile, 'r') as ff:
            for li, line in enumerate([xx.strip() for xx in
                                       ff.read().splitlines()]):
                if line.startswith('(') and li <= len(radiosourcedict):
                    key = line.split()[0]
                    bibc = line.split()[-1]
                    radiosourcedict[key] = catalog.entries[
                        name].add_source(bibcode=bibc)
                elif li in [xx + len(radiosourcedict) for xx in range(3)]:
                    continue
                else:
                    cols = list(filter(None, line.split()))
                    source = radiosourcedict[cols[6]]
                    catalog.entries[name].add_photometry(
                        time=cols[0], frequency=cols[2], u_frequency='GHz',
                        fluxdensity=cols[3], e_fluxdensity=cols[4],
                        u_fluxdensity='µJy',
                        instrument=cols[5], source=source)
                    catalog.entries[name].add_quantity(
                        'alias', oldname, source)

    catalog.journal_entries()
    return


def do_external_xray(catalog):
    task_str = catalog.get_current_task_str()
    path_pattern = os.path.join(catalog.get_current_task_repo(), '*.txt')
    for datafile in pbar_strings(glob(path_pattern), task_str):
        oldname = os.path.basename(datafile).split('.')[0]
        name = catalog.add_entry(oldname)
        with open(datafile, 'r') as ff:
            for li, line in enumerate(ff.read().splitlines()):
                if li == 0:
                    source = catalog.entries[name].add_source(
                        bibcode=line.split()[-1])
                elif li in [1, 2, 3]:
                    continue
                else:
                    cols = list(filter(None, line.split()))
                    catalog.entries[name].add_photometry(
                        time=cols[:2],
                        energy=cols[2:4], u_energy='keV', counts=cols[4],
                        flux=cols[6],
                        unabsorbedflux=cols[8], u_flux='ergs/ss/cm^2',
                        photonindex=cols[15], instrument=cols[
                            17], nhmw=cols[11],
                        upperlimit=(float(cols[5]) < 0), source=source)
                    catalog.entries[name].add_quantity(
                        'alias', oldname, source)

    catalog.journal_entries()
    return


def do_internal(catalog):
    """Load events from files in the 'internal' repository, and save them.
    """
    task_str = catalog.get_current_task_str()
    path_pattern = os.path.join(catalog.get_current_task_repo(), '*.json')
    files = glob(path_pattern)
    catalog.log.debug("found {} files matching '{}'".format(
        len(files), path_pattern))
    for datafile in pbar_strings(files, task_str):
        new_event = Supernova.init_from_file(
            catalog, path=datafile, clean=True)
        catalog.entries.update({new_event[KEYS.NAME]: new_event})

    return



"""
"""
from astrocats.catalog.utils import is_number


class KeyCollection:
    """General container class with methods to list attribute names and values.

    Used mostly by different `CatDict` subclasses to contain the 'keys' to
    their internal dictionaries.
    """
    _keys = []
    _vals = []

    @classmethod
    def keys(cls):
        """Return this class's attribute names (those not stating with '_').

        Returns
        -------
        _keys : list of str
            List of names of internal attributes.  Order is effectiely random.
        """
        if cls._keys:
            return cls._keys

        cls._keys = [kk for kk in vars(cls).keys() if not kk.startswith('_')]
        return cls._keys

    @classmethod
    def vals(cls):
        """Return this class's attribute values (those not stating with '_').

        Returns
        -------
        _vals : list of objects
            List of values of internal attributes.  Order is effectiely random,
            but should match that returned by `keys()`.
        """
        if cls._vals:
            return cls._vals

        cls._vals = [vv for kk, vv in
                     vars(cls).items() if not kk.startswith('_')]
        return cls._vals


class KEY_TYPES(KeyCollection):
    NUMERIC = 'numeric'
    STRING = 'string'
    BOOL = 'bool'
    ANY = None


class Key(str):
    """Class to act as a 'key' (with metadata) to a `CatDict` dictionary.

    Used in `KeyCollection` subclasses, for example `PHOTOMETRY`. The most
    basic function is to contain the str 'key' for a `CatDict` (or other)
    dictionary, but can also contain (and manage) additional specifictions and
    metadata. For example, if a `type` parameter is specified (from
    `KEY_TYPES`), then the builtin `check()` method will ensure that a value is
    consistent with that type.  Arbitrary additional attributes can also be
    stored.

    Attributes
    ----------
    name : str
        The str used as a key for the corresponding dictionary.
    type : str or None (one of `KEY_TYPES`)
        Specification of variable type for the value corresponding to this key.
        'None' means any type is valid.
        NOTE: if ``type == None``, then the `listable` check is disabled.
        note: if `type` is not one of `KEY_TYPES`, a `ValueError` is raised.
    listable : bool
        Whether the value corresponding to this key is allowed to be a list.
        NOTE: this is only enforced if `type` is not 'None'.
    compare : bool
        Whether the value corresponding to this key should be compared to that
        of other dictionaries when looking for duplicate entries.
        i.e. is equality of the value corresponding to this key requisite for
        a different dictionary to be considered a duplicate.

    """
    def __new__(cls, name, type=None, listable=False, compare=True, **kwargs):
        """Construct the underlying str object with `name`.
        """
        return str.__new__(cls, name)

    def __init__(self, name, type=None,
                 listable=False, compare=True, **kwargs):
        # Make sure type is allowed
        if type is not None and type not in KEY_TYPES.vals():
            raise ValueError(
                "Key `type` ('{}') must be 'None' or one of '{}'".format(
                    type, KEY_TYPES.keys()))
        self.name = str(name)
        self.type = type
        self.listable = listable
        self.compare = compare
        for key, val in kwargs.items():
            setattr(self, key, val)

    def pretty(self):
        """Return a 'pretty' string representation of this `Key`.

        note: do not override the builtin `__str__` or `__repr__` methods!
        """
        retval = "Key(name={}, type={}, listable={}, compare={})".format(
            self.name, self.type, self.listable, self.compare)
        return retval

    def check(self, val):
        """Make sure given value is consistent with this `Key` specification.

        NOTE: if `type` is 'None', then `listable` also is *not* checked.
        """
        # If there is no `type` requirement, everything is allowed
        if self.type is None:
            return True

        is_list = isinstance(val, list)
        # If lists are not allowed, and this is a list --> false
        if not self.listable and is_list:
            return False

        # `is_number` already checks for either list or single value
        if self.type == KEY_TYPES.NUMERIC and not is_number(val):
            return False
        elif self.type == KEY_TYPES.STRING:
            # If its a list, check first element
            if is_list and not isinstance(val[0], str):
                return False
            # Otherwise, check it
            elif not isinstance(val, str):
                return False
        elif self.type == KEY_TYPES.BOOL:
            if is_list and not isinstance(val[0], bool):
                return False
            elif not isinstance(val, bool):
                return False

        return True



"""Import tasks for the Dark Energy Survey.
"""
import json
import os

from bs4 import BeautifulSoup

from astrocats.catalog.utils import pbar


def do_des(catalog):
    task_str = catalog.get_current_task_str()
    des_url = 'https://portal.nersc.gov/des-sn/'
    des_trans_url = des_url + 'transients/'
    ackn_url = ('http://www.noao.edu/'
                'noao/library/NOAO_Publications_Acknowledgments.html'
                '#DESdatause')
    # Make sure there is aa trailing slash
    des_path = os.path.join(catalog.get_current_task_repo(), 'DES', '')
    html = catalog.load_cached_url(des_trans_url, des_path + 'transients.html')
    if not html:
        return
    bs = BeautifulSoup(html, 'html5lib')
    trs = bs.find('tbody').findAll('tr')
    for tri, tr in enumerate(pbar(trs, task_str)):
        name = ''
        # source = ''
        if tri == 0:
            continue
        tds = tr.findAll('td')
        for tdi, td in enumerate(tds):
            if tdi == 0:
                name = catalog.add_entry(td.text.strip())
            if tdi == 1:
                (ra, dec) = [xx.strip() for xx in td.text.split('\xa0')]
            if tdi == 6:
                atellink = td.find('a')
                if atellink:
                    atellink = atellink['href']
                else:
                    atellink = ''

        sources = [catalog.entries[name]
                   .add_source(url=des_url, name='DES Bright Transients',
                               acknowledgment=ackn_url)]
        if atellink:
            sources.append(
                catalog.entries[name]
                .add_source(name='ATel ' + atellink.split('=')[-1],
                            url=atellink))
        sources += [catalog.entries[name]
                    .add_source(bibcode='2012ApJ...753..152B'),
                    catalog.entries[name].add_source(
                        bibcode='2015AJ....150..150F'),
                    catalog.entries[name].add_source(
                        bibcode='2015AJ....150...82G'),
                    catalog.entries[name]
                    .add_source(bibcode='2015AJ....150..172K')]
        sources = ','.join(sources)
        catalog.entries[name].add_quantity('alias', name, sources)
        catalog.entries[name].add_quantity('ra', ra, sources)
        catalog.entries[name].add_quantity('dec', dec, sources)

        html2 = catalog.load_cached_url(
            des_trans_url + name, des_path + name + '.html')
        if not html2:
            continue
        lines = html2.splitlines()
        for line in lines:
            if 'var data = ' in line:
                jsontxt = json.loads(line.split('=')[-1].rstrip(';'))
                for ii, band in enumerate(jsontxt['band']):
                    upl = True if float(jsontxt['snr'][ii]) <= 3.0 else ''
                    (catalog.entries[name]
                     .add_photometry(time=jsontxt['mjd'][ii],
                                     magnitude=jsontxt['mag'][ii],
                                     e_magnitude=jsontxt['mag_error'][ii],
                                     band=band, observatory='CTIO',
                                     telescope='Blanco 4m', instrument='DECam',
                                     upperlimit=upl, source=sources))

    catalog.journal_entries()
    return



'''Clean various supernova-specific values.
'''
from astrocats.catalog.utils import is_number, zpad, get_sig_digits, pretty_num
from cdecimal import Decimal
from math import floor

__all__ = ['name_clean', 'host_clean', 'radec_clean', 'clean_snname']


def name_clean(name):
    newname = name.strip(' ;,*')
    if newname.startswith('NAME '):
        newname = newname.replace('NAME ', '', 1)
    if newname.endswith(' SN'):
        newname = newname.replace(' SN', '')
    if newname.endswith(':SN'):
        newname = newname.replace(':SN', '')
    if newname.startswith('MASJ'):
        newname = newname.replace('MASJ', 'MASTER OT J', 1)
    if newname.startswith('MASTER') and is_number(newname[7]):
        newname = newname.replace('MASTER', 'MASTER OT J', 1)
    if newname.startswith('MASTER OT J '):
        newname = newname.replace('MASTER OT J ', 'MASTER OT J', 1)
    if newname.startswith('OGLE '):
        newname = newname.replace('OGLE ', 'OGLE-', 1)
    if newname.startswith('OGLE-') and len(newname) != 16:
        namesp = newname.split('-')
        if (len(namesp[1]) == 4 and is_number(namesp[1]) and
                is_number(namesp[3])):
            newname = 'OGLE-' + namesp[1] + '-SN-' + namesp[3].zfill(3)
    if newname.startswith('SN SDSS'):
        newname = newname.replace('SN SDSS ', 'SDSS', 1)
    if newname.startswith('SDSS '):
        newname = newname.replace('SDSS ', 'SDSS', 1)
    if newname.startswith('SDSS'):
        namesp = newname.split('-')
        if (len(namesp) == 3 and is_number(namesp[0][4:]) and
                is_number(namesp[1]) and is_number(namesp[2])):
            newname = namesp[0] + '-' + namesp[1] + '-' + namesp[2].zfill(3)
    if newname.startswith('SDSS-II SN'):
        namesp = newname.split()
        if len(namesp) == 3 and is_number(namesp[2]):
            newname = 'SDSS-II SN ' + namesp[2].lstrip('0')
    if newname.startswith('SN CL'):
        newname = newname.replace('SN CL', 'CL', 1)
    if newname.startswith('SN HiTS '):
        newname = newname.replace('SN HiTS ', 'SNHiTS', 1)
    if newname.startswith('GAIA'):
        newname = newname.replace('GAIA', 'Gaia', 1)
    if newname.startswith('Gaia '):
        newname = newname.replace('Gaia ', 'Gaia', 1)
    if newname.startswith('Gaia'):
        newname = 'Gaia' + newname[4:].lower()
    if newname.startswith('GRB'):
        newname = newname.replace('GRB', 'GRB ', 1)
    if newname.startswith('GRB ') and is_number(newname[4:].strip()):
        newname = 'GRB ' + newname[4:].strip() + 'A'
    if newname.startswith('LSQ '):
        newname = newname.replace('LSQ ', 'LSQ', 1)
    if newname.startswith('KSN '):
        newname = newname.replace('KSN ', 'KSN-', 1)
    if newname.startswith('SNSDF '):
        newname = newname.replace(' ', '')
    if newname.startswith('SNSDF'):
        namesp = newname.split('.')
        if len(namesp[0]) == 9:
            newname = namesp[0] + '-' + namesp[1].zfill(2)
    if newname.startswith('HFF '):
        newname = newname.replace(' ', '')
    if newname.startswith('SN HST'):
        newname = newname.replace('SN HST', 'HST', 1)
    if newname.startswith('HST ') and newname[4] != 'J':
        newname = newname.replace('HST ', 'HST J', 1)
    if newname.startswith('SNLS') and newname[4] != '-':
        newname = newname.replace('SNLS', 'SNLS-', 1)
    if newname.startswith('SNLS- '):
        newname = newname.replace('SNLS- ', 'SNLS-', 1)
    if newname.startswith('CRTS CSS'):
        newname = newname.replace('CRTS CSS', 'CSS', 1)
    if newname.startswith('CRTS MLS'):
        newname = newname.replace('CRTS MLS', 'MLS', 1)
    if newname.startswith('CRTS SSS'):
        newname = newname.replace('CRTS SSS', 'SSS', 1)
    if newname.startswith(('CSS', 'MLS', 'SSS')):
        newname = newname.replace(' ', ':').replace('J', '')
    if newname.startswith('SN HFF'):
        newname = newname.replace('SN HFF', 'HFF', 1)
    if newname.startswith('SN GND'):
        newname = newname.replace('SN GND', 'GND', 1)
    if newname.startswith('SN SCP'):
        newname = newname.replace('SN SCP', 'SCP', 1)
    if newname.startswith('SN UDS'):
        newname = newname.replace('SN UDS', 'UDS', 1)
    if newname.startswith('SCP') and newname[3] != '-':
        newname = newname.replace('SCP', 'SCP-', 1)
    if newname.startswith('SCP- '):
        newname = newname.replace('SCP- ', 'SCP-', 1)
    if newname.startswith('PS 1'):
        newname = newname.replace('PS 1', 'PS1', 1)
    if newname.startswith('PS1 SN PS'):
        newname = newname.replace('PS1 SN PS', 'PS', 1)
    if newname.startswith('PS1 SN'):
        newname = newname.replace('PS1 SN', 'PS1', 1)
    if newname.startswith('PSN K'):
        newname = newname.replace('PSN K', 'K', 1)
    if newname.startswith('K') and is_number(newname[1:5]):
        namesp = newname.split('-')
        if len(namesp[0]) == 5:
            newname = namesp[0] + '-' + namesp[1].zfill(3)
    if newname.startswith('Psn'):
        newname = newname.replace('Psn', 'PSN', 1)
    if newname.startswith('PSNJ'):
        newname = newname.replace('PSNJ', 'PSN J', 1)
    if newname.startswith('TCPJ'):
        newname = newname.replace('TCPJ', 'TCP J', 1)
    if newname.startswith('SMTJ'):
        newname = newname.replace('SMTJ', 'SMT J', 1)
    if newname.startswith('PSN20J'):
        newname = newname.replace('PSN20J', 'PSN J', 1)
    if newname.startswith('SN ASASSN'):
        newname = newname.replace('SN ASASSN', 'ASASSN', 1)
    if newname.startswith('ASASSN '):
        newname = newname.replace('ASASSN ', 'ASASSN-', 1).replace('--', '-')
    if newname.startswith('ASASSN') and newname[6] != '-':
        newname = newname.replace('ASASSN', 'ASASSN-', 1)
    if newname.startswith('ROTSE3J'):
        newname = newname.replace('ROTSE3J', 'ROTSE3 J', 1)
    if newname.startswith('MACSJ'):
        newname = newname.replace('MACSJ', 'MACS J', 1)
    if newname.startswith('MWSNR'):
        newname = newname.replace('MWSNR', 'MWSNR ', 1)
    if newname.startswith('SN HUNT'):
        newname = newname.replace('SN HUNT', 'SNhunt', 1)
    if newname.startswith('SN Hunt'):
        newname = newname.replace(' ', '')
    if newname.startswith('SNHunt'):
        newname = newname.replace('SNHunt', 'SNhunt', 1)
    if newname.startswith('SNhunt '):
        newname = newname.replace('SNhunt ', 'SNhunt', 1)
    if newname.startswith('ptf'):
        newname = newname.replace('ptf', 'PTF', 1)
    if newname.startswith('SN PTF'):
        newname = newname.replace('SN PTF', 'PTF', 1)
    if newname.startswith('PTF '):
        newname = newname.replace('PTF ', 'PTF', 1)
    if newname.startswith('iPTF '):
        newname = newname.replace('iPTF ', 'iPTF', 1)
    if newname.startswith('PESSTOESO'):
        newname = newname.replace('PESSTOESO', 'PESSTO ESO ', 1)
    if newname.startswith('snf'):
        newname = newname.replace('snf', 'SNF', 1)
    if newname.startswith('SNF '):
        newname = newname.replace('SNF ', 'SNF', 1)
    if (newname.startswith('SNF') and
            is_number(newname[3:]) and len(newname) >= 12):
        newname = 'SNF' + newname[3:11] + '-' + newname[11:]
    if newname.startswith(('MASTER OT J', 'ROTSE3 J')):
        prefix = newname.split('J')[0]
        coords = newname.split('J')[-1].strip()
        decsign = '+' if '+' in coords else '-'
        coordsplit = coords.replace('+', '-').split('-')
        if ('.' not in coordsplit[0] and
                len(coordsplit[0]) > 6 and '.' not in coordsplit[1] and
                len(coordsplit[1]) > 6):
            newname = (prefix + 'J' + coordsplit[0][:6] + '.' +
                       coordsplit[0][6:] + decsign + coordsplit[1][:6] +
                       '.' + coordsplit[1][6:])
    if (newname.startswith('Gaia ') and
            is_number(newname[3:4]) and len(newname) > 5):
        newname = newname.replace('Gaia ', 'Gaia', 1)
    if len(newname) <= 4 and is_number(newname):
        newname = 'SN' + newname + 'A'
    if (len(newname) > 4 and is_number(newname[:4]) and not
            is_number(newname[4:])):
        newname = 'SN' + newname
    if (newname.startswith('Sn ') and
            is_number(newname[3:7]) and len(newname) > 7):
        newname = newname.replace('Sn ', 'SN', 1)
    if (newname.startswith('sn') and
            is_number(newname[2:6]) and len(newname) > 6):
        newname = newname.replace('sn', 'SN', 1)
    if (newname.startswith('SN ') and
            is_number(newname[3:7]) and len(newname) > 7):
        newname = newname.replace('SN ', 'SN', 1)
    if (newname.startswith('SN') and
            is_number(newname[2:6]) and len(newname) == 7 and
            newname[6].islower()):
        newname = 'SN' + newname[2:6] + newname[6].upper()
    elif (newname.startswith('SN') and is_number(newname[2:6]) and
          (len(newname) == 8 or len(newname) == 9) and newname[6:].isupper()):
        newname = 'SN' + newname[2:6] + newname[6:].lower()

    newname = (' '.join(newname.split())).strip()
    return newname


def radec_clean(svalue, quantity, unit=''):
    if unit == 'floatdegrees':
        if not is_number(svalue):
            return (svalue, unit)
        deg = float('%g' % Decimal(svalue))
        sig = get_sig_digits(svalue)
        if 'ra' in quantity:
            flhours = deg / 360.0 * 24.0
            hours = floor(flhours)
            minutes = floor((flhours - hours) * 60.0)
            seconds = (flhours * 60.0 - (hours * 60.0 + minutes)) * 60.0
            hours = 0 if hours < 1.e-6 else hours
            minutes = 0 if minutes < 1.e-6 else minutes
            seconds = 0.0 if seconds < 1.e-6 else seconds
            if seconds > 60.0:
                raise(ValueError('Invalid seconds value for ' + quantity))
            svalue = str(hours).zfill(2) + ':' + str(minutes).zfill(2) + \
                ':' + zpad(pretty_num(seconds, sig=sig - 1))
        elif 'dec' in quantity:
            fldeg = abs(deg)
            degree = floor(fldeg)
            minutes = floor((fldeg - degree) * 60.0)
            seconds = (fldeg * 60.0 - (degree * 60.0 + minutes)) * 60.0
            if seconds > 60.0:
                raise(ValueError('Invalid seconds value for ' + quantity))
            svalue = (('+' if deg >= 0.0 else '-') +
                      str(degree).strip('+-').zfill(2) + ':' +
                      str(minutes).zfill(2) + ':' +
                      zpad(pretty_num(seconds, sig=sig - 1)))
    elif unit == 'nospace' and 'ra' in quantity:
        svalue = svalue[:2] + ':' + svalue[2:4] + \
            ((':' + zpad(svalue[4:])) if len(svalue) > 4 else '')
    elif unit == 'nospace' and 'dec' in quantity:
        if svalue.startswith(('+', '-')):
            svalue = svalue[:3] + ':' + svalue[3:5] + \
                ((':' + zpad(svalue[5:])) if len(svalue) > 5 else '')
        else:
            svalue = '+' + svalue[:2] + ':' + svalue[2:4] + \
                ((':' + zpad(svalue[4:])) if len(svalue) > 4 else '')
    else:
        svalue = svalue.replace(' ', ':')
        if 'dec' in quantity:
            valuesplit = svalue.split(':')
            svalue = (('-' if valuesplit[0].startswith('-') else '+') +
                      valuesplit[0].strip('+-').zfill(2) +
                      (':' + valuesplit[1].zfill(2) if
                       len(valuesplit) > 1 else '') +
                      (':' + zpad(valuesplit[2]) if
                       len(valuesplit) > 2 else ''))

    if 'ra' in quantity:
        sunit = 'hours'
    elif 'dec' in quantity:
        sunit = 'degrees'

    # Correct case of arcseconds = 60.0.
    valuesplit = svalue.split(':')
    if len(valuesplit) == 3 and valuesplit[-1] in ["60.0", "60.", "60"]:
        svalue = valuesplit[0] + ':' + str(Decimal(valuesplit[1]) +
                                           Decimal(1.0)) + ':' + "00.0"

    # Strip trailing dots.
    svalue = svalue.rstrip('.')

    return (svalue, sunit)


def host_clean(name):
    newname = name.strip(' ;,*')

    # Handle some special cases
    hostcases = {'M051a': 'M51A', 'M051b': 'M51B'}
    for k in hostcases:
        if newname == k:
            newname = hostcases[k]

    # Some general cases
    newname = newname.strip("()").replace('  ', ' ', 1)
    newname = newname.replace("ABELL", "Abell", 1)
    newname = newname.replace("Abell", "Abell ", 1)
    newname = newname.replace("APMUKS(BJ)", "APMUKS(BJ) ", 1)
    newname = newname.replace("ARP", "ARP ", 1)
    newname = newname.replace("CGCG", "CGCG ", 1)
    newname = newname.replace("HOLM", "HOLM ", 1)
    newname = newname.replace("IC", "IC ", 1)
    newname = newname.replace("Intergal.", "Intergalactic", 1)
    newname = newname.replace("MCG+", "MCG +", 1)
    newname = newname.replace("MCG-", "MCG -", 1)
    newname = newname.replace("M+", "MCG +", 1)
    newname = newname.replace("M-", "MCG -", 1)
    newname = newname.replace("MGC ", "MCG ", 1)
    newname = newname.replace("Mrk", "MRK", 1)
    newname = newname.replace("MRK", "MRK ", 1)
    newname = newname.replace("NGC", "NGC ", 1)
    newname = newname.replace("PGC", "PGC ", 1)
    newname = newname.replace("SDSS", "SDSS ", 1)
    newname = newname.replace("UGC", "UGC ", 1)
    if newname.startswith('MESSIER '):
        newname = newname.replace('MESSIER ', 'M', 1)
    if newname.startswith('M ') and is_number(newname[2:]):
        newname = newname.replace('M ', 'M', 1)
    if newname.startswith('M') and is_number(newname[1:]):
        newname = 'M' + newname[1:].lstrip(" 0")
    if len(newname) > 4 and newname.startswith("PGC "):
        newname = newname[:4] + newname[4:].lstrip(" 0")
    if len(newname) > 4 and newname.startswith("UGC "):
        newname = newname[:4] + newname[4:].lstrip(" 0")
    if len(newname) > 5 and newname.startswith(("MCG +", "MCG -")):
        newname = newname[:5] + '-'.join([x.zfill(2)
                                          for x in
                                          newname[5:].strip().split("-")])
    if len(newname) > 5 and newname.startswith("CGCG "):
        newname = newname[:5] + '-'.join([x.zfill(3)
                                          for x in
                                          newname[5:].strip().split("-")])
    if ((len(newname) > 1 and newname.startswith("E")) or
            (len(newname) > 3 and newname.startswith('ESO'))):
        if newname[0] == "E":
            esplit = newname[1:].split("-")
        else:
            esplit = newname[3:].split("-")
        if len(esplit) == 2 and is_number(esplit[0].strip()):
            if esplit[1].strip()[0] == 'G':
                parttwo = esplit[1][1:].strip()
            else:
                parttwo = esplit[1].strip()
            if is_number(parttwo.strip()):
                newname = 'ESO ' + \
                    esplit[0].lstrip('0') + '-G' + parttwo.lstrip('0')
    newname = ' '.join(newname.split())
    return newname


def clean_snname(string):
    newstring = string.replace(' ', '').upper()
    if (newstring[:2] == "SN"):
        head = newstring[:6]
        tail = newstring[6:]
        if len(tail) >= 2 and tail[1] != '?':
            tail = tail.lower()
        newstring = head + tail

    return newstring



"""Import tasks for the Transient Name Server.
"""
import csv
import os
from math import ceil

import requests
from astropy.time import Time as astrotime

from astrocats.catalog.utils import pbar, pretty_num


def do_tns(catalog):
    from datetime import timedelta
    session = requests.Session()
    task_str = catalog.get_current_task_str()
    tns_url = 'https://wis-tns.weizmann.ac.il/'
    search_url = tns_url + \
        'search?&num_page=1&format=html&sort=desc&order=id&format=csv&page=0'
    csvtxt = catalog.load_cached_url(
        search_url, os.path.join(catalog.get_current_task_repo(),
                                 'TNS/index.csv'))
    if not csvtxt:
        return
    maxid = csvtxt.splitlines()[1].split(',')[0].strip('"')
    maxpages = ceil(int(maxid) / 1000.)

    for page in pbar(range(maxpages), task_str):
        fname = os.path.join(catalog.get_current_task_repo(), 'TNS/page-') + \
            str(page).zfill(2) + '.csv'
        if (catalog.current_task.load_archive(catalog.args) and
                os.path.isfile(fname) and page < 7):
            with open(fname, 'r') as tns_file:
                csvtxt = tns_file.read()
        else:
            with open(fname, 'w') as tns_file:
                session = requests.Session()
                ses_url = (tns_url + 'search?&num_page=1000&format=html&edit'
                           '[type]=&edit[objname]=&edit[id]=&sort=asc&order=id'
                           '&display[redshift]=1'
                           '&display[hostname]=1&display[host_redshift]=1'
                           '&display[source_group_name]=1'
                           '&display[programs_name]=1'
                           '&display[internal_name]=1'
                           '&display[isTNS_AT]=1'
                           '&display[public]=1'
                           '&display[end_pop_period]=0'
                           '&display[spectra_count]=1'
                           '&display[discoverymag]=1&display[discmagfilter]=1'
                           '&display[discoverydate]=1&display[discoverer]=1'
                           '&display[sources]=1'
                           '&display[bibcode]=1&format=csv&page=' + str(page))
                try:
                    response = session.get(ses_url)
                    csvtxt = response.text
                except:
                    if os.path.isfile(fname):
                        with open(fname, 'r') as tns_file:
                            csvtxt = tns_file.read()
                    else:
                        continue
                else:
                    tns_file.write(csvtxt)

        tsvin = list(csv.reader(csvtxt.splitlines(), delimiter=','))
        for ri, row in enumerate(pbar(tsvin, task_str, leave=False)):
            if ri == 0:
                continue
            if row[4] and 'SN' not in row[4]:
                continue
            name = row[1].replace(' ', '')
            name = catalog.add_entry(name)
            source = catalog.entries[name].add_source(
                name='Transient Name Server', url=tns_url)
            catalog.entries[name].add_quantity('alias', name, source)
            if row[2] and row[2] != '00:00:00.00':
                catalog.entries[name].add_quantity('ra', row[2], source)
            if row[3] and row[3] != '+00:00:00.00':
                catalog.entries[name].add_quantity('dec', row[3], source)
            if row[4]:
                catalog.entries[name].add_quantity(
                    'claimedtype', row[4].replace('SN', '').strip(), source)
            if row[5]:
                catalog.entries[name].add_quantity(
                    'redshift', row[5], source, kind='spectroscopic')
            if row[6]:
                catalog.entries[name].add_quantity('host', row[6], source)
            if row[7]:
                catalog.entries[name].add_quantity(
                    'redshift', row[7], source, kind='host')
            if row[8]:
                catalog.entries[name].add_quantity(
                    'discoverer', row[8], source)
            # Currently, all events listing all possible observers. TNS bug?
            # if row[9]:
            #    observers = row[9].split(',')
            #    for observer in observers:
            #        catalog.entries[name].add_quantity('observer',
            #                                  observer.strip(),
            #                                  source)
            if row[10]:
                catalog.entries[name].add_quantity('alias', row[10], source)
            if row[8] and row[14] and row[15] and row[16]:
                survey = row[8]
                magnitude = row[14]
                band = row[15].split('-')[0]
                mjd = astrotime(row[16]).mjd
                (catalog.entries[name]
                 .add_photometry(time=mjd, magnitude=magnitude,
                                 band=band, survey=survey, source=source))
            if row[16]:
                date = row[16].split()[0].replace('-', '/')
                if date != '0000/00/00':
                    date = date.replace('/00', '')
                    time = row[16].split()[1]
                    if time != '00:00:00':
                        ts = time.split(':')
                        dt = timedelta(hours=int(ts[0]), minutes=int(
                            ts[1]), seconds=int(ts[2]))
                        date += pretty_num(dt.total_seconds() /
                                           (24 * 60 * 60), sig=6).lstrip('0')
                    catalog.entries[name].add_quantity(
                        'discoverdate', date, source)
            if catalog.args.update:
                catalog.journal_entries()

    catalog.journal_entries()
    return



"""
"""
import statistics
import warnings
from math import hypot, log10, pi, sqrt

from astropy import units as un
from astropy.coordinates import SkyCoord as coord
from astropy.cosmology import Planck15 as cosmo
from astropy.cosmology import z_at_value

from astrocats.catalog.utils import (get_sig_digits, is_number, pbar,
                                     pretty_num, tprint, uniq_cdl)
from astrocats.supernovae.constants import CLIGHT, KM, PREF_KINDS
from cdecimal import Decimal


def do_derivations(catalog):
    task_str = catalog.get_current_task_str()
    # Calculate some columns based on imported data, sanitize some fields
    for oname in pbar(catalog.entries, task_str):
        name = catalog.add_entry(oname)

        aliases = catalog.entries[name].get_aliases()
        catalog.entries[name].set_first_max_light()

        if 'discoverdate' not in catalog.entries[name]:
            prefixes = ['MLS', 'SSS', 'CSS', 'GRB ']
            for alias in aliases:
                for prefix in prefixes:
                    if (alias.startswith(prefix) and
                            is_number(alias.replace(prefix, '')[:2])):
                        discoverdate = ('/'.
                                        join(['20' +
                                              alias.replace(prefix, '')[:2],
                                              alias.replace(prefix, '')[2:4],
                                              alias.replace(prefix, '')[4:6]]))
                        if catalog.args.verbose:
                            tprint(
                                'Added discoverdate from name [' +
                                alias + ']: ' + discoverdate)
                        source = catalog.entries[name].add_self_source()
                        catalog.entries[name].add_quantity(
                            'discoverdate', discoverdate, source, derived=True)
                        break
                if 'discoverdate' in catalog.entries[name]:
                    break
        if 'discoverdate' not in catalog.entries[name]:
            prefixes = ['ASASSN-', 'PS1-', 'PS1', 'PS', 'iPTF', 'PTF', 'SCP-',
                        'SNLS-', 'SPIRITS', 'LSQ', 'DES', 'SNHiTS', 'Gaia',
                        'GND', 'GNW', 'GSD', 'GSW', 'EGS', 'COS', 'OGLE',
                        'HST']
            for alias in aliases:
                for prefix in prefixes:
                    if (alias.startswith(prefix) and
                            is_number(alias.replace(prefix, '')[:2]) and
                            is_number(alias.replace(prefix, '')[:1])):
                        discoverdate = '20' + alias.replace(prefix, '')[:2]
                        if catalog.args.verbose:
                            tprint(
                                'Added discoverdate from name [' +
                                alias + ']: ' + discoverdate)
                        source = catalog.entries[name].add_self_source()
                        catalog.entries[name].add_quantity(
                            'discoverdate', discoverdate, source, derived=True)
                        break
                if 'discoverdate' in catalog.entries[name]:
                    break
        if 'discoverdate' not in catalog.entries[name]:
            prefixes = ['SNF']
            for alias in aliases:
                for prefix in prefixes:
                    if (alias.startswith(prefix) and
                            is_number(alias.replace(prefix, '')[:4])):
                        discoverdate = ('/'
                                        .join(
                                            [alias.replace(prefix, '')[:4],
                                             alias.replace(prefix, '')[4:6],
                                             alias.replace(prefix, '')[6:8]]))
                        if catalog.args.verbose:
                            tprint(
                                'Added discoverdate from name [' +
                                alias + ']: ' + discoverdate)
                        source = catalog.entries[name].add_self_source()
                        catalog.entries[name].add_quantity(
                            'discoverdate', discoverdate, source, derived=True)
                        break
                if 'discoverdate' in catalog.entries[name]:
                    break
        if 'discoverdate' not in catalog.entries[name]:
            prefixes = ['PTFS', 'SNSDF']
            for alias in aliases:
                for prefix in prefixes:
                    if (alias.startswith(prefix) and
                            is_number(alias.replace(prefix, '')[:2])):
                        discoverdate = ('/'
                                        .join(
                                            ['20' +
                                             alias.replace(prefix, '')[:2],
                                             alias.replace(prefix, '')[2:4]]))
                        if catalog.args.verbose:
                            tprint(
                                'Added discoverdate from name [' +
                                alias + ']: ' + discoverdate)
                        source = catalog.entries[name].add_self_source()
                        catalog.entries[name].add_quantity(
                            'discoverdate', discoverdate, source, derived=True)
                        break
                if 'discoverdate' in catalog.entries[name]:
                    break
        if 'discoverdate' not in catalog.entries[name]:
            prefixes = ['AT', 'SN', 'OGLE-', 'SM ', 'KSN-']
            for alias in aliases:
                for prefix in prefixes:
                    if (alias.startswith(prefix) and
                            is_number(alias.replace(prefix, '')[:4]) and
                            '.' not in alias.replace(prefix, '')[:4]):
                        discoverdate = alias.replace(prefix, '')[:4]
                        if catalog.args.verbose:
                            tprint(
                                'Added discoverdate from name [' +
                                alias + ']: ' + discoverdate)
                        source = catalog.entries[name].add_self_source()
                        catalog.entries[name].add_quantity(
                            'discoverdate', discoverdate, source, derived=True)
                        break
                if 'discoverdate' in catalog.entries[name]:
                    break

        if ('ra' not in catalog.entries[name] or
                'dec' not in catalog.entries[name]):
            prefixes = ['PSN J', 'MASJ', 'CSS', 'SSS', 'MASTER OT J', 'HST J',
                        'TCP J', 'MACS J', '2MASS J', 'EQ J', 'CRTS J',
                        'SMT J']
            for alias in aliases:
                for prefix in prefixes:
                    if (alias.startswith(prefix) and
                            is_number(alias.replace(prefix, '')[:6])):
                        noprefix = alias.split(
                            ':')[-1].replace(prefix, '').replace('.', '')
                        decsign = '+' if '+' in noprefix else '-'
                        noprefix = noprefix.replace('+', '|').replace('-', '|')
                        nops = noprefix.split('|')
                        if len(nops) < 2:
                            continue
                        rastr = nops[0]
                        decstr = nops[1]
                        ra = ':'.join([rastr[:2], rastr[2:4], rastr[4:6]]) + \
                            ('.' + rastr[6:] if len(rastr) > 6 else '')
                        dec = (decsign + ':'
                               .join([decstr[:2], decstr[2:4], decstr[4:6]]) +
                               ('.' + decstr[6:] if len(decstr) > 6 else ''))
                        if catalog.args.verbose:
                            tprint('Added ra/dec from name: ' + ra + ' ' + dec)
                        source = catalog.entries[name].add_self_source()
                        catalog.entries[name].add_quantity(
                            'ra', ra, source, derived=True)
                        catalog.entries[name].add_quantity(
                            'dec', dec, source, derived=True)
                        break
                if 'ra' in catalog.entries[name]:
                    break

        no_host = ('host' not in catalog.entries[name] or
                   not any([x['value'] == 'Milky Way' for x in
                            catalog.entries[name]['host']]))
        if ('ra' in catalog.entries[name] and
                'dec' in catalog.entries[name] and no_host):
            from astroquery.irsa_dust import IrsaDust
            if name not in catalog.extinctions_dict:
                try:
                    ra_dec = catalog.entries[name]['ra'][0]['value'] + \
                        " " + catalog.entries[name]['dec'][0]['value']
                    result = IrsaDust.get_query_table(ra_dec, section='ebv')
                except (KeyboardInterrupt, SystemExit):
                    raise
                except:
                    warnings.warn("Coordinate lookup for " +
                                  name + " failed in IRSA.")
                else:
                    ebv = result['ext SandF mean'][0]
                    ebverr = result['ext SandF std'][0]
                    catalog.extinctions_dict[name] = [ebv, ebverr]
            if name in catalog.extinctions_dict:
                sources = uniq_cdl(
                    [catalog.entries[name].add_self_source(),
                     catalog.entries[name]
                     .add_source(bibcode='2011ApJ...737..103S')])
                (catalog.entries[name]
                 .add_quantity('ebv',
                               str(catalog
                                   .extinctions_dict[name][0]),
                               sources,
                               error=str(catalog
                                         .extinctions_dict[name][1]),
                               derived=True))
        if (('host' in catalog.entries[name] and
             ('hostra' not in catalog.entries[name] or
              'hostdec' not in catalog.entries[name]))):
            for host in catalog.entries[name]['host']:
                alias = host['value']
                if ' J' in alias and is_number(alias.split(' J')[-1][:6]):
                    noprefix = alias.split(
                        ' J')[-1].split(':')[-1].replace('.', '')
                    decsign = '+' if '+' in noprefix else '-'
                    noprefix = noprefix.replace('+', '|').replace('-', '|')
                    nops = noprefix.split('|')
                    if len(nops) < 2:
                        continue
                    rastr = nops[0]
                    decstr = nops[1]
                    hostra = (':'.join([rastr[:2], rastr[2:4], rastr[4:6]]) +
                              ('.' + rastr[6:] if len(rastr) > 6 else ''))
                    hostdec = decsign + ':'.join([decstr[:2], decstr[2:4],
                                                  decstr[4:6]]) + (
                        '.' + decstr[6:] if len(decstr) > 6 else '')
                    if catalog.args.verbose:
                        tprint('Added hostra/hostdec from name: ' +
                               hostra + ' ' + hostdec)
                    source = catalog.entries[name].add_self_source()
                    catalog.entries[name].add_quantity(
                        'hostra', hostra, source, derived=True)
                    catalog.entries[name].add_quantity(
                        'hostdec', hostdec, source, derived=True)
                    break
                if 'hostra' in catalog.entries[name]:
                    break

        if ('redshift' not in catalog.entries[name] and
                'velocity' in catalog.entries[name]):
            # Find the "best" velocity to use for this
            bestsig = 0
            for hv in catalog.entries[name]['velocity']:
                sig = get_sig_digits(hv['value'])
                if sig > bestsig:
                    besthv = hv['value']
                    bestsrc = hv['source']
                    bestsig = sig
            if bestsig > 0 and is_number(besthv):
                voc = float(besthv) * 1.e5 / CLIGHT
                source = catalog.entries[name].add_self_source()
                sources = uniq_cdl([source] + bestsrc.split(','))
                (catalog.entries[name]
                 .add_quantity('redshift',
                               pretty_num(sqrt((1. + voc) / (1. - voc)) - 1.,
                                          sig=bestsig),
                               sources, kind='heliocentric',
                               derived=True))
        if ('redshift' not in catalog.entries[name] and
                len(catalog.nedd_dict) > 0 and
                'host' in catalog.entries[name]):
            reference = "NED-D"
            refurl = "http://ned.ipac.caltech.edu/Library/Distances/"
            for host in catalog.entries[name]['host']:
                if host['value'] in catalog.nedd_dict:
                    source = catalog.entries[name].add_source(
                        bibcode='2015arXiv150201589P')
                    secondarysource = catalog.entries[name].add_source(
                        name=reference, url=refurl, secondary=True)
                    meddist = statistics.median(
                        catalog.nedd_dict[host['value']])
                    redz = z_at_value(
                        cosmo.comoving_distance, float(meddist) * un.Mpc)
                    redshift = pretty_num(
                        redz, sig=get_sig_digits(str(meddist)))
                    catalog.entries[name].add_quantity(
                        name, 'redshift', redshift,
                        uniq_cdl([source, secondarysource]),
                        kind='host', derived=True)
        if ('maxabsmag' not in catalog.entries[name] and
                'maxappmag' in catalog.entries[name] and
                'lumdist' in catalog.entries[name]):
            # Find the "best" distance to use for this
            bestsig = 0
            for ld in catalog.entries[name]['lumdist']:
                sig = get_sig_digits(ld['value'])
                if sig > bestsig:
                    bestld = ld['value']
                    bestsrc = ld['source']
                    bestsig = sig
            if bestsig > 0 and is_number(bestld) and float(bestld) > 0.:
                source = catalog.entries[name].add_self_source()
                sources = uniq_cdl([source] + bestsrc.split(','))
                # FIX: what's happening here?!
                pnum = (float(catalog.entries[name]['maxappmag'][0]['value']) -
                        5.0 * (log10(float(bestld) * 1.0e6) - 1.0))
                pnum = pretty_num(pnum, sig=bestsig)
                catalog.entries[name].add_quantity(
                    'maxabsmag', pnum, sources, derived=True)
        if 'redshift' in catalog.entries[name]:
            # Find the "best" redshift to use for this
            bestz, bestkind, bestsig, bestsrc = catalog.entries[
                name].get_best_redshift()
            if bestsig > 0:
                bestz = float(bestz)
                if 'velocity' not in catalog.entries[name]:
                    source = catalog.entries[name].add_self_source()
                    # FIX: what's happening here?!
                    pnum = CLIGHT / KM * \
                        ((bestz + 1.)**2. - 1.) / ((bestz + 1.)**2. + 1.)
                    pnum = pretty_num(pnum, sig=bestsig)
                    catalog.entries[name].add_quantity(
                        'velocity', pnum, source, kind=PREF_KINDS[bestkind])
                if bestz > 0.:
                    from astropy.cosmology import Planck15 as cosmo
                    if 'lumdist' not in catalog.entries[name]:
                        dl = cosmo.luminosity_distance(bestz)
                        sources = [
                            catalog.entries[name].add_self_source(),
                            catalog.entries[name]
                            .add_source(bibcode='2015arXiv150201589P')]
                        sources = uniq_cdl(sources + bestsrc.split(','))
                        catalog.entries[name].add_quantity(
                            'lumdist', pretty_num(dl.value, sig=bestsig),
                            sources, kind=PREF_KINDS[bestkind],
                            derived=True)
                        if ('maxabsmag' not in catalog.entries[name] and
                                'maxappmag' in catalog.entries[name]):
                            source = catalog.entries[name].add_self_source()
                            pnum = pretty_num(
                                float(catalog.entries[name]['maxappmag'][0][
                                    'value']) -
                                5.0 * (log10(dl.to('pc').value) - 1.0),
                                sig=bestsig)
                            catalog.entries[name].add_quantity(
                                'maxabsmag', pnum, sources, derived=True)
                    if 'comovingdist' not in catalog.entries[name]:
                        cd = cosmo.comoving_distance(bestz)
                        sources = [
                            catalog.entries[name].add_self_source(),
                            catalog.entries[name]
                            .add_source(bibcode='2015arXiv150201589P')]
                        sources = uniq_cdl(sources + bestsrc.split(','))
                        catalog.entries[name].add_quantity(
                            'comovingdist', pretty_num(cd.value, sig=bestsig),
                            sources, derived=True)
        if all([x in catalog.entries[name] for x in
                ['ra', 'dec', 'hostra', 'hostdec']]):
            # For now just using first coordinates that appear in entry
            try:
                c1 = coord(
                    ra=catalog.entries[name]['ra'][0][
                        'value'], dec=catalog.entries[name]['dec'][0]['value'],
                    unit=(un.hourangle, un.deg))
                c2 = coord(
                    ra=catalog.entries[name]['hostra'][0]['value'],
                    dec=catalog.entries[name]['hostdec'][0]['value'],
                    unit=(un.hourangle, un.deg))
            except (KeyboardInterrupt, SystemExit):
                raise
            except:
                pass
            else:
                sources = uniq_cdl(
                    [catalog.entries[name].add_self_source()] +
                    catalog.entries[name]['ra'][0]['source'].split(',') +
                    catalog.entries[name]['dec'][0]['source'].split(',') +
                    catalog.entries[name]['hostra'][0]['source'].split(',') +
                    catalog.entries[name]['hostdec'][0]['source'].split(','))
                if 'hostoffsetang' not in catalog.entries[name]:
                    hosa = Decimal(hypot(c1.ra.degree - c2.ra.degree,
                                         c1.dec.degree - c2.dec.degree))
                    hosa = pretty_num(hosa * Decimal(3600.))
                    catalog.entries[name].add_quantity(
                        'hostoffsetang', hosa, sources,
                        derived=True, unit='arcseconds')
                if ('comovingdist' in catalog.entries[name] and
                        'redshift' in catalog.entries[name] and
                        'hostoffsetdist' not in catalog.entries[name]):
                    offsetsig = get_sig_digits(
                        catalog.entries[name]['hostoffsetang'][0]['value'])
                    sources = uniq_cdl(sources.split(',') +
                                       (catalog.entries[name]['comovingdist']
                                        [0]['source']).split(',') +
                                       (catalog.entries[name]['redshift']
                                        [0]['source']).split(','))
                    (catalog.entries[name]
                     .add_quantity('hostoffsetdist',
                                   pretty_num(
                                       float(catalog.entries[name][
                                           'hostoffsetang']
                                           [0]['value']) /
                                       3600. * (pi / 180.) *
                                       float(catalog.entries[name][
                                           'comovingdist']
                                           [0]['value']) *
                                       1000. / (1.0 +
                                                float(catalog.entries[name][
                                                    'redshift']
                                                    [0]['value'])),
                                       sig=offsetsig), sources))

        catalog.journal_entries()

    return



"""
"""
import os

from astrocats.supernovae.supernova import KEYS

TEST_NAME = 'SN-TEST-AA'

FAKE_BIBCODE_1 = '2025Tst...123..456Z'
FAKE_ALIAS_1 = 'PS-TEST-AB'
FAKE_REDZ_1 = '1.123'

FAKE_BIBCODE_2 = '1925Tst...987..654A'
FAKE_ALIAS_2 = 'PTF-TEST-BA'
FAKE_REDZ_2 = '0.987'


def do_test(catalog):
    log = catalog.log
    log.error("do_test()")
    task_str = catalog.get_current_task_str()
    log.error("`task_str`: '{}'".format(task_str))

    if len(catalog.entries) != 0:
        raise RuntimeError("Run test only with empty catalog.")

    # Create a Fake Entry, with some Fake Data
    # ----------------------------------------
    _first_source(catalog)

    #
    log_str = "ADDING SECOND SOURCE"
    log.error("\n\n{}\n{}\n{}\n\n".format("=" * 100, log_str, "=" * 100))

    # Add new Data, from different source, to same fake entry
    # -------------------------------------------------------
    _second_source(catalog)

    # Make sure output file for this test exists
    outdir, filename = catalog.entries[TEST_NAME]._get_save_path()
    save_name = os.path.join(outdir, filename + '.json')
    if not os.path.exists(save_name):
        raise RuntimeError("File not found in '{}'".format(save_name))
    # Delete created test file
    catalog._delete_entry_file(entry_name=TEST_NAME)
    # Make sure it was deleted
    if os.path.exists(save_name):
        raise RuntimeError("File not deleted at '{}'".format(save_name))

    # Delete entry in catalog
    del catalog.entries[TEST_NAME]
    # Make sure entry was deleted
    if len(catalog.entries) != 0:
        raise RuntimeError("Error deleting test entry!")

    return


def _first_source(catalog):
    """Try adding a single source, with some data.
    """
    log = catalog.log
    # Add Entry to Catalog
    log.error("Calling: ``add_entry('{}')``".format(TEST_NAME))
    name = catalog.add_entry(TEST_NAME)
    log.error("\t `name`: '{}'".format(name))
    log.error("\n{}\n".format(repr(catalog.entries[name])))
    # Make sure entry exists
    if TEST_NAME not in catalog.entries:
        raise RuntimeError("`TEST_NAME`: '{}' is not in entries".format(
            TEST_NAME))
    # Make sure entry has the correct name
    stored_name = catalog.entries[TEST_NAME][KEYS.NAME]
    if stored_name != TEST_NAME:
        raise RuntimeError("`TEST_NAME`[{}]: '{}' does not match".format(
            KEYS.NAME, stored_name, TEST_NAME))

    # Add source to entry
    log.error("Calling: ``add_source('{}')``".format(FAKE_BIBCODE_1))
    source = catalog.entries[name].add_source(
        bibcode=FAKE_BIBCODE_1)
    log.error("\t `source`: '{}'".format(source))
    log.error("\n{}\n".format(repr(catalog.entries[name])))
    # Make sure source alias is correct
    if source != '1':
        raise RuntimeError("Returned `source`: '{}' is wrong.".format(source))
    # Make sure source has the right properties
    check_source_1(catalog, name)

    # Add alias
    log.error("Calling: ``add_quantity('alias', '{}', '{}')``".format(
        FAKE_ALIAS_1, source))
    catalog.entries[name].add_quantity('alias', FAKE_ALIAS_1, source)
    log.error("\n{}\n".format(repr(catalog.entries[name])))
    # Make sure source alias is correct
    stored_aliases = catalog.entries[name][KEYS.ALIAS]
    if ((len(stored_aliases) != 1 or
         stored_aliases[0]['value'] != FAKE_ALIAS_1 or
         stored_aliases[0]['source'] != source)):
        raise RuntimeError("Stored alias: '{}' looks wrong.".format(
            stored_aliases[0]))

    log.error("Calling: ``add_quantity('redshift', '{}', '{}')``".format(
        FAKE_REDZ_1, source))
    catalog.entries[name].add_quantity(
        'redshift', FAKE_REDZ_1, source, kind='spectroscopic')
    log.error("\n{}\n".format(repr(catalog.entries[name])))

    log.error("Calling: ``journal_entries()``")
    catalog.journal_entries()
    log.error("\n{}\n".format(repr(catalog.entries[name])))
    # Make sure the remaining stub looks right
    check_stub(catalog, name)
    return


def _second_source(catalog):
    log = catalog.log

    log.error("Calling: ``add_entry('{}')``".format(FAKE_ALIAS_1))
    name = catalog.add_entry(FAKE_ALIAS_1)
    log.error("\t `name`: '{}'".format(name))
    log.error("\n{}\n".format(repr(catalog.entries[name])))
    # Make sure the proper name is returned (instead of the alias)
    if name != TEST_NAME:
        raise RuntimeError("Returned `name`: '{}' does not match '{}'".format(
            name, TEST_NAME))
    # Make sure previous data was loaded
    check_source_1(catalog, name)

    log.error("Calling: ``add_source('{}')``".format(FAKE_BIBCODE_2))
    source = catalog.entries[name].add_source(
        bibcode=FAKE_BIBCODE_2)
    log.error("\t `source`: '{}'".format(source))
    log.error("\n{}\n".format(repr(catalog.entries[name])))

    log.error("Calling: ``add_quantity('alias', '{}', '{}')``".format(
        FAKE_ALIAS_2, source))
    catalog.entries[name].add_quantity('alias', FAKE_ALIAS_2, source)
    log.error("\n{}\n".format(repr(catalog.entries[name])))
    check_source_2(catalog, name)

    log.error("Calling: ``add_quantity('redshift', '{}', '{}')``".format(
        FAKE_REDZ_2, source))
    catalog.entries[name].add_quantity(
        'redshift', FAKE_REDZ_2, source, kind='spectroscopic')
    log.error("\n{}\n".format(repr(catalog.entries[name])))

    log.error("Calling: ``journal_entries()``")
    catalog.journal_entries()
    log.error("\n{}\n".format(repr(catalog.entries[name])))
    check_stub(catalog, name)
    return


def check_source_1(catalog, name):
    stored_sources = catalog.entries[name][KEYS.SOURCES]
    if ((len(stored_sources) != 1 or
         stored_sources[0][KEYS.NAME] != FAKE_BIBCODE_1 or
         stored_sources[0][KEYS.BIBCODE] != FAKE_BIBCODE_1)):
        raise RuntimeError("Stored source: '{}' looks wrong.".format(
            stored_sources[0]))
    return


def check_source_2(catalog, name):
    stored_sources = catalog.entries[name][KEYS.SOURCES]
    names = [src[KEYS.NAME] for src in stored_sources]
    codes = [src[KEYS.NAME] for src in stored_sources]
    if ((len(stored_sources) != 2 or
         FAKE_BIBCODE_1 not in names or FAKE_BIBCODE_2 not in names or
         FAKE_BIBCODE_1 not in codes or FAKE_BIBCODE_2 not in codes)):
        raise RuntimeError("Stored sources: '{}' look wrong.".format(
            stored_sources))
    return


def check_stub(catalog, name):
    if not catalog.entries[name]._stub:
        raise RuntimeError("Remaining entry is not a stub.")
    if KEYS.ALIAS not in catalog.entries[name]:
        raise RuntimeError("Remaining entry is missing '{}'.".format(
            KEYS.ALIAS))
    if KEYS.SOURCES in catalog.entries[name]:
        raise RuntimeError("Remaining still has '{}'.".format(
            KEYS.SOURCES))
    return
